/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* VER: $Id: LinearPredictor.c,v 1.4 2002/11/04 02:46:18 olethros Exp $ */
/* LinearPredictor - Uses a weight sum to predict the next value of a time-series.  */


#include "LinearPredictor.h"
//==========================================================
// NewLinearPredictor
//----------------------------------------------------------
//
// Note that the weight array is organized as follows:
//
// There is a weight set for each of the d outputs,
// which includes l sets of d-dimensional vectors,
// one for each sample in time.

LinearPredictor* NewLinearPredictor(int l, int d)
{
  int i = 0;
  LinearPredictor* pred = NULL;

  pred = AllocM (LinearPredictor, 1);
  
  if (!pred) return NULL;
  
  pred->n = l;
  pred->d = d;
  if ((pred->n > 0) && (pred->d > 0)) {
    pred->w = AllocM (real, d*l*d);
    pred->error = AllocM (real, d);
    pred->out = AllocM (real, d);
  }else{
    pred->w = NULL;
    pred->n = 0;
    pred->d = 0;
    Swarning("%d inputs and %d dimensions specified - no allocation done\n",l,d);
  }

  if (! ((pred->error) || (pred->out) || (pred->w))) {
    goto quick_exit;
  }
  
  pred->a = 0.01f;
  for (i=0; i<d; i++) {
    pred->out[i] = 0.0f;
    pred->error[i] = 0.0f;
  }
  
  return pred;

 quick_exit:
  fprintf(stderr, "Allocation errror!");
  if (pred->w) {
    FreeM(pred->w);
  }

  if (pred->error) {
    FreeM(pred->error);
  }

  if (pred->out) {
    FreeM(pred->out);
  }

  FreeM(pred);
  return NULL;
}


//==========================================================
// DeleteLinearPredictor
//----------------------------------------------------------

int DeleteLinearPredictor(LinearPredictor* pred)
{
  assert(pred);
  if (pred->n) {
    if (pred->w) {
      FreeM(pred->w);
    } else {
      Serror("Null weight vector when freeing predictor\n");
      return -1;
    }
    if (pred->out) {
      FreeM(pred->out);
    }else{
      Serror("Null output vector when freeing predictor\n");
      return -1;
    }
    if (pred->error) {
      FreeM(pred->error);
    }else{
      Serror("Null error vector when freeing predictor\n");
      return -1;
    }
  }else{
    Swarning("This predictor had 0 inputs.");
  }

  return 0;
}

//==========================================================
// LinearPredictor_Reset - Initializes the parameters
//----------------------------------------------------------
void LinearPredictor_Reset(LinearPredictor* pred)
{
  long i;
  for (i=0; i<((pred->d)*(pred->n)*(pred->d)); i++) {
    pred->w[i] = 0.0f;
  }
}


//==========================================================
// LinearPredictor_Input - Give an input
//----------------------------------------------------------
real* LinearPredictor_Input(LinearPredictor* pred, real* x)
{
  long i;
  long j;
  long k;
  real* w = pred->w;
  real* y = x;

  //  fprintf(stderr,"[Input]\n");
  for (j=0; j<pred->d; j++) {
    real sum  = 0.0f;
    y = x;
    for (i=0; i<pred->n; i++) {
      for (k=0; k<pred->d; k++) {
        sum += (*w++)*(*y++);
      }
    }
    pred->out[j] = sum;
  }

  return pred->out;
}


//==========================================================
// LinearPredictor - train on a sequence
//----------------------------------------------------------
// Note that this particular overloaded version assumes that
// the target output is x[n] and the inputs are x[0]..x[n-1]
real* LinearPredictor_Train(LinearPredictor* pred, real* x)
{
  real* error = LinearPredictor_Input(pred, x);
  real* y = &x[(pred->n)*(pred->d)]; // should be the last element
  long i;

  assert(y==LinearPredictor_GetMatrixElement(pred, x, pred->n, 0));

  for (i=0;i<pred->d;i++) {
    pred->error[i] = *error++ - *y++;
  }
  LinearPredictor_Gradient(pred, x, pred->error);

  return pred->error;
}



//==========================================================
// LinearPredictor_Test()
//----------------------------------------------------------
real LinearPredictor_Test(LinearPredictor* pred, real* x)
{
  real* error = LinearPredictor_Input(pred, x);
  real* y = &x[(pred->n)*(pred->d)]; // should be the last element
  long i;
  
  assert(y==LinearPredictor_GetMatrixElement(pred, x, pred->n, 0));

  for (i=0;i<pred->d;i++) {
    pred->error[i] = *error++ - *y++;
  }

  return LinearPredictor_GetError(pred);
}


//==========================================================
// LinearPredictor_Gradient()
//----------------------------------------------------------
void LinearPredictor_Gradient(LinearPredictor* pred, real* x, real* e)
{
  long i,j, k;
  real* w = pred->w;
  real* y;
  real a;

  w = pred->w;

  for (j=0; j<pred->d; j++) {
    a = (pred->a)* e[j];
    y = x;
    for (i=0; i<pred->n; i++) {
      for (k=0; k<pred->d; k++) {
        *w++ -= *(y++)*a;
      }
    }
  }

}

//==========================================================
// LinearPredictor
//----------------------------------------------------------
real* LinearPredictor_GetOutput(LinearPredictor* pred)
{
  return pred->out;
}
//==========================================================
// LinearPredictor
//----------------------------------------------------------
real LinearPredictor_GetError(LinearPredictor* pred)
{
  long i;
  real x = 0.0f;
  for (i=0;i<pred->d;i++) {
    register real y  = pred->error[i];
    x = y*y;
  }
  
  return sqrt(x);
}

//==========================================================
// LinearPredictor_SetLearningRate
//----------------------------------------------------------
void LinearPredictor_SetLearningRate(LinearPredictor* pred, real a)
{
  pred->a = a;
}

//==========================================================
// LinearPredictor_GetMatrixElement
//----------------------------------------------------------
// Get the address of a particular element in a single-pointer matrix
real* LinearPredictor_GetMatrixElement(LinearPredictor* pred, real* x, long time, long dimension)
{
  return (x + (time*pred->d + dimension));
}
