/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* VER: $Id: LinearPredictor.h,v 1.7 2003/07/25 03:19:00 olethros Exp $ */

#ifndef LINEAR_PREDICTOR_H
#define LINEAR_PREDICTOR_H

#include <math.h>
#include "debug.h"
#include "util.h"
#include "string_utils.h"
#include "LinkedList.h"
#include "datafile.h"
#include "real.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>


//==========================================================
// LinearPredictor
//----------------------------------------------------------
//
// Input format:
// - The n-th vector element of an N-dimensional input at time t
// is $x_{t,n}, n\in[0,N-1]$
//
// In the code, it is impleneted as
// x_{t,n} def: x[t*N + n];
//
// This means that vector components for a particular time t can be
// addressed sequentially.
// 
// Weight format:
// - Weights are arranged in the exact same manner.
//
// To make things simpler, a simple function that returns a pointer
// to the element that you want to look at is implemented. However,
// for speed reasons, it is not used in the inner loops.
// 
// SEE ALSO
// - LinearPredictor.c : For an explanation of all functions

typedef struct LinearPredictor_ {
  long n; // predictor time-window length
  long d; // dimension (number of channels)
  real* w; // weights
  real* out; // output
  real a; // learning rate
  real* error;
} LinearPredictor;

//--- User level functions ---//
// Creation and deletion of predictors
LinearPredictor* NewLinearPredictor(int l, int d);
int DeleteLinearPredictor(LinearPredictor* pred);

// Initialise predictor weights
void LinearPredictor_Reset(LinearPredictor* pred);

// Set the learning rate
void LinearPredictor_SetLearningRate(LinearPredictor* pred, real a);

// Train for single-step prediction on the given input
real* LinearPredictor_Train(LinearPredictor* pred, real* x);

// Get the single-step prediction error norm for the given input
real LinearPredictor_Test(LinearPredictor* pred, real* x);


//--- Medium level functions ---//

// Perform gradient descent according to the given error vector
void LinearPredictor_Gradient(LinearPredictor* pred, real* x, real* e);

// Do a forward pass through the network
real* LinearPredictor_Input(LinearPredictor* pred, real* x);

// Get the network's current output
real* LinearPredictor_GetOutput(LinearPredictor* pred);

// Get the network's current error norm
real LinearPredictor_GetError(LinearPredictor* pred);

//--- Low level functions ---//

// Get the address of a particular element in a single-pointer matrix
real* LinearPredictor_GetMatrixElement(LinearPredictor* pred, real* x, long time, long dimension);

#endif /* LINEAR_PREDICTOR_H */
