/* -*- Mode: C++;  -*- */
/* VER: $Id: Predictor.h,v 1.7 2003/09/05 01:08:43 olethros Exp $ */
#ifndef PREDICTOR_H

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "real.h"
#include "debug.h"
#include "util.h"
#include "ANN.h"
#include "Windowing.h"

/*
 * Given an n-dimensional input sequence S, where n_inputs is the
 * number of dimensions, the predictor attempts to learn a mapping
 * from $d$ delayed intputs, where d is the embedding dimension to
 * T+look_ahead outputs. Each input x_i is delayed by an amount i*T,
 * so x[i] = S[t-i*T] .* Each output y_j is a prediction of
 * S[t+j]. For a look_ahead of 0, the predictor simply tries to fill
 * in the gaps between S[t-T] and S[t]. 
 *
 */
typedef struct Predictor_ {
	int n_inputs; // dimensionality of the sequence
	int d; // embedding dimension
	int T; // delay
	int look_ahead; // look ahead (additional)
	ANN* ann;
	SampleWindow* windowing_function;
	real* x; // The compacted inputs
	real* y; // The compacted target outputs
	void* (*NewPredictor) (int l, int d);
	void* (*DeletePredictor) (void* pred);
	void* (*Reset) (void* pred);
	void* (*SetLearningRate) (void* pred, real a);
	real* (*Train) (void* pred, real* x);
	real (*Test) (void* pred, real* x);
	void* (*Gradient) (void* pred, real* x, real* e);
	real* (*Input) (void* pred, real* x);
	real* (*GetOutput) (void* pred);
	real (*GetError) (void* pred);
	real* (*GetMatrixElement) (void* pred, real* x, long time, long dimension);
} Predictor;

Predictor* NewPredictor (int n_inputs, int d, int T, int look_ahead);
/* Create a predictor using a neural network */
Predictor* NewANNPredictor (int n_inputs, int d, int T, int look_ahead);
/* Create a predictor with hidden control neural network */
Predictor* NewHCNNPredictor (int n_inputs, int d, int T, int look_ahead);
int DeletePredictor (Predictor* pred);
/* Set up */
int Predictor_AddHiddenLayer (Predictor* pred, int n_units);
int Predictor_AddRBFHiddenLayer (Predictor* pred, int n_units);
void Predictor_SetLearningRate (Predictor* pred, real a);
void Predictor_SetBatchMode(Predictor*, bool batch);
int Predictor_Init (Predictor* pred);
void Predictor_SetOutputsToTanH (Predictor* pred);
void Predictor_SetWindowingFunction (SampleWindow* window);
/* Functionality Interface */
real Predictor_Input(Predictor* predictor, real* x);
real Predictor_Train(Predictor* predictor, real* x);
void Predictor_BatchAdapt(Predictor* pred);
real Predictor_Test(Predictor* predictor, real* x);
real* Predictor_GetOutput(Predictor* predictor);
real Predictor_GetError(Predictor* predictor);
real Predictor_HCNNInput(Predictor* predictor, real* x, real c);
real Predictor_HCNNTrain(Predictor* predictor, real* x, real c);

/* Low-level */
void Predictor_CompactInputs(Predictor* predictor, real* x);
void Predictor_HCNNCompactInputs(Predictor* predictor, real* x, real c);

#endif /* Predictor_H */
