/* -*- Mode: C++; -*- */
/*! \brief Single Input Multiple Output buffer.  
 * 
 * This class implemets single input multiple output buffers. 
 *
 */

#ifndef SIMO_BUFFER_H
#define SIMO_BUFFER_H

#include "IOPort.h"
#include "SmartAssert.h"

template <typename T>
class SIMOBuffer {
public:
	SIMOBuffer (int outputs, int size); ///< Create a buffer with size
	~SIMOBuffer ();
	IOPort<T>* WritePort_Req ();
	IOPort<T>* ReadPort_Req();
	bool IO_Req (IOPort<T>* port);
private:
	void Acknowledge();
	int n_outputs;
	T* buffer; ///< holds the data buffer
	T* buffer_end;
	int length; ///< length
	IOPort<T>* output_port; 
	IOPort<T>* input_port;
	T* first_write_pos;
	T* last_write_pos;
	T* first_read_pos;
	T* last_read_pos;
	int open_out_ports;
	int n_ack;
};

template <typename T>
SIMOBuffer<T>::SIMOBuffer (int outputs, int size)
{
	n_outputs = outputs;
	length = size;
	buffer = new T[length];
	buffer_end = &buffer[length];
	first_write_pos = buffer;
	last_write_pos = buffer;
	first_read_pos = buffer;
	last_read_pos = buffer;

	input_port = new IOPort<T>;
	input_port->flags = None;
	input_port->head = buffer;
	input_port->tail = buffer;
	input_port->Ack = NULL;
	input_port->ACK_HOLD = 0;

	open_out_ports = 0;
	output_port = new IOPort<T> [n_outputs];
	for (int i=0; i<n_outputs; i++) {
		output_port[i].flags = None;
		output_port[i].head = buffer;
		output_port[i].tail = buffer;
		output_port[i].Ack = NULL;
		output_port[i].ACK_HOLD = 0;
	}
	n_ack = 0;
}

template <typename T>
SIMOBuffer<T>::~SIMOBuffer ()
{
	delete [] output_port;
	delete input_port;
	delete [] buffer;
}

template <typename T>
IOPort<T>* SIMOBuffer<T>::WritePort_Req()
{
	if (input_port->flags) {
		return NULL;
	}
	input_port->flags = Write;

	return input_port;
}

template <typename T>
IOPort<T>* SIMOBuffer<T>::ReadPort_Req()
{
	SMART_ASSERT (open_out_ports<n_outputs)(open_out_ports)(n_outputs).error("No more available slots");
	SMART_ASSERT (output_port[open_out_ports].flags==None)(output_port[open_out_ports].flags).error("Port already open");
	IOPort<T>* r= &output_port[open_out_ports++];
	r->flags= Read;
	return r;
}

/// Request to perform IO.
/** If the port is open, true is returned, otherwise false.  Note that
	this is a HOLD signal, so multiple invocations have the same
	effect as a single invocation.
*/
template <typename T>
bool SIMOBuffer<T>::IO_Req (IOPort<T>* port)
{
	switch (port->flags) {
	case None:
		return false;
	case Read:
		if (port->head < first_read_pos) {
			first_read_pos = port->head;
		}
		if (port->tail > last_read_pos) {
			last_read_pos = port->tail;
		}
		break;
	case Write:
		if (port->head < first_write_pos) {
			first_write_pos = port->head;
		}
		if (port->tail > last_write_pos) {
			last_write_pos = port->tail;
		}
	default:
		SMART_ASSERT (true)(port->flags).error ("Invalid flags");
	}
	if (port->ACK_HOLD==false) {
		port->ACK_HOLD = true;
		n_ack++;
	}
	Acknowledge();

	return true;
}


/// After we are full of requests, acknowledge to everybody
template <typename T>
void SIMOBuffer<T>::Acknowledge()
{
	T* read_head = buffer_end;
	T* read_tail = buffer;
	for (int i=0; i<open_out_ports; i++) {
//		if (output_port[i].ACK_HOLD) {
			if (output_port[i].head < read_head) {
				read_head = output_port[i].head;
			}
			if (output_port[i].tail > read_tail) {
				read_tail = output_port[i].tail;
			}
//		}
	}
	std::cout << "Moving read head from " << (int) (first_read_pos-buffer) << " to " << (int) (read_head-buffer) << std::endl;
	std::cout << "Moving read tail from " << (int) (last_read_pos-buffer) << " to " << (int) (read_tail-buffer) << std::endl;
	first_read_pos = read_head;
	last_read_pos = read_tail;

	// Check the write lock.
	if (first_write_pos < first_read_pos) {
		first_write_pos = first_read_pos;
	}

	if (input_port->ACK_HOLD) {
		// Boundary condition, move everything as much left as possible.
		// Note we must also copy the data
		if (input_port->tail >= buffer_end) {
			std::cout << "Moving writing head to the beginning" << std::endl;
			// First we must see where is the last write.
			int diff_pos;
			if (first_read_pos > input_port->head) {
				diff_pos = input_port->head - buffer;
			} else {
				diff_pos = first_read_pos - buffer;
			}

			input_port->tail -=  diff_pos;
			input_port->head -= diff_pos;
			for (int i=0; i<open_out_ports; i++) {
				output_port->head -= diff_pos;
				output_port->tail -= diff_pos;
			}
			first_write_pos -= diff_pos;
			last_write_pos -= diff_pos;
			first_read_pos -= diff_pos;
			last_read_pos -= diff_pos;
		}
		last_write_pos = input_port->head;
		input_port->ACK_HOLD = false;
		n_ack--;
		input_port->Ack (input_port);

	}

	// OK, Only satisfy reader requests that are within range
	// we only check head> and tail < as head < tail should hold.
	for (int i=0; i<open_out_ports; i++) {
		SMART_ASSERT (output_port[i].head <= output_port[i].tail)(i).error ("Mismathed head and tail for output port.");
		if ((output_port[i].head >= first_write_pos) && 
			(output_port[i].tail < last_write_pos)) {
			output_port[i].ACK_HOLD=false;
			n_ack--;
			output_port[i].Ack (&output_port[i]);
		}
	}
}


#endif
