/* -*- Mode: C++; -*- */
/* "VER: $Id: SampleBuffer.c,v 1.19 2003/09/05 01:08:43 olethros Exp $" */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "debug.h"
#include "SampleBuffer.h"
#include "util.h"



//==========================================================
// NewSampleBuffer()
//----------------------------------------------------------
/*
  DESCRIPTION

  - Creates a new sample buffer of the specified length.

  ARGUMENTS

  - length : The length of the buffer, in samples.
  - flags : 0, or CLEAR_SAMPLE_BUFFER to clear the buffer
  after creation.

  RETURN VALUE

  - A pointer to the buffer if the operation was successful, otherwise NULL.

  SEE ALSO

  - LoadSampleBuffer(), FreeSampleBuffer()
*/
SAMPLE_BUFFER* NewSampleBuffer(long length, int flags) {
	SAMPLE_BUFFER* buffer;
  
	if (length<0) {
		return NULL;
	}
	buffer = (SAMPLE_BUFFER*) malloc(sizeof(SAMPLE_BUFFER));

	if (buffer==NULL) {
		return NULL;
	}

	buffer->length = length;  
	if (length) {
		if ((buffer->x = (float*) malloc(length*sizeof(float)))==NULL) {
			free (buffer);
			fprintf (stderr, "Could not allocate sample buffer! (size %ld too large)\n", length);
			return NULL;
		}

		if (flags & CLEAR_SAMPLE_BUFFER) {
			ClearBuffer (buffer);
		}
	} else {
		buffer->x = NULL; //else just an empty buffer
	}

	return buffer;

} /* NewSampleBuffer() */



//==========================================================
// SetBufferLength()
//----------------------------------------------------------
/*
  DESCRIPTION

  - Set the buffer length to a particular value 

  ARGUMENTS

  - length : The length of the buffer, in samples.

  RETURN VALUE

  - 0 if the operation was sucessful.

  SEE ALSO

  - NewSampleBuffer(), RequestBufferLength(), FreeSampleBuffer()
*/

long SetBufferLength (SAMPLE_BUFFER* buffer, long length)
{
	if (buffer->x) {
		if (buffer->length == length) {
			return length;
		}
		free (buffer->x);
	}

	if ((buffer->x = (float*) malloc(length*sizeof(float)))==NULL) {
		Serror ("Could not set requested size");
		buffer->length = 0;
		return 0;
	}
	buffer->length = length;
	return length;
}

//==========================================================
// RequestBufferLength()
//----------------------------------------------------------
/*
  DESCRIPTION

  - Request a minimum buffer length 

  ARGUMENTS

  - length : The length of the buffer, in samples.

  RETURN VALUE

  - 0 if the operation was sucessful.

  SEE ALSO

  - NewSampleBuffer(), SetBufferLength(), FreeSampleBuffer()
*/
long RequestBufferLength (SAMPLE_BUFFER* buffer, long length)
{
	if (buffer->x) {
		if (buffer->length >= length) {
			return buffer->length;
		}
		free (buffer->x);
	}

	if ((buffer->x = (float*) malloc(length*sizeof(float)))==NULL) {
		Serror ("Could not set requested size");
		buffer->length = 0;
		return 0;
	}
	buffer->length = length;
	return length;
}


void ClearBuffer (SAMPLE_BUFFER* buffer) 
{
    register long i;
    register float* x = buffer->x;
    register long length = buffer->length;
    for (i=0;i<length;i++) {
		*x++ = 0.0f;
    }
}

//==========================================================
// LoadSampleBuffer()
//----------------------------------------------------------
/*
  DESCRIPTION

  - Creates a new sample buffer from a file, assuming the data are
  stored in 16-bit (2==sizeof(short)?) binary signed format, with
  endianness the same as the machine's native. This only works for
  raw data. Look at SndFileSampleBuffer() for a version using
  libsndfile.

  ARGUMENTS

  - fname : The name of the file to load the buffer from.  

  RETURN VALUE

  - A pointer to the buffer if the operation was successful, otherwise
  NULL.

  SEE ALSO

  - LoadASCIISampleBuffer(), FreeSampleBuffer(), SndFileSampleBuffer();
*/
SAMPLE_BUFFER* LoadSampleBuffer(const char* fname) {
	SAMPLE_BUFFER* buffer = NULL;
	FILE* f;
	long l;
	float* x;
	float scale;

  
	l = FileLength(fname) >> 1;

	if (ReadFile(fname, f)) {    
		long i;
		short* tbuff = NULL;
		tbuff = AllocM(short, l);
		if (!tbuff) return NULL;
		buffer = NewSampleBuffer(l, 0);
		if (!buffer) {
			FreeM(tbuff);
			return NULL;
		}

		fread(tbuff, sizeof(short), l, f); 
		x = buffer->x;
		assert(l==buffer->length);
		scale = 1.0f/32768.0f;
		for (i=0; i<l; i++) {
			*x++ = ((float) tbuff[i])*scale;
		}

		CloseFile(f);
		FreeM(tbuff);

		return buffer;
	} 

  
	return NULL;
}
//==========================================================
// SndFileSampleBuffer()
//----------------------------------------------------------
/*
  DESCRIPTION

  - Creates a new sample buffer from a file, using libsndfile.

  ARGUMENTS

  - fname : The name of the file to load the buffer from.  

  RETURN VALUE

  - A pointer to the buffer if the operation was successful, otherwise
  NULL.

  SEE ALSO

  - LoadASCIISampleBuffer(), FreeSampleBuffer(), LoadSampleBuffer();
*/
SAMPLE_BUFFER* SndFileSampleBuffer (const char* fname)
{
	SNDFILE* tmp = NULL;
	SF_INFO file_info;
	SAMPLE_BUFFER* data = NULL;
	tmp = sf_open (fname, SFM_READ, &file_info);
	if (!(sf_format_check(&file_info))) {
		Serror ("Invalid format?");
		sf_close (tmp);
		return NULL;
	}
  
	data = NewSampleBuffer (file_info.frames, 0);
	data->rate = file_info.samplerate;
	sf_read_float (tmp, data->x, file_info.frames);
	sf_close(tmp);

	return data;
}

//==========================================================
// LoadASCIISampleBuffer()
//----------------------------------------------------------
/*
  DESCRIPTION

  - Creates a new sample buffer from a file, assuming the data are in
  ASCII numerical format.

  ARGUMENTS

  - fname : THe name of the file to load the buffer from.

  RETURN VALUE

  - A pointer to the buffer if the operation was successful, otherwise
  NULL.

  SEE ALSO

  - LoadSampleBuffer(), FreeSampleBuffer()
*/
SAMPLE_BUFFER* LoadASCIISampleBuffer(const char* fname) {
	SAMPLE_BUFFER* buffer = NULL;
	FILE* f;
	float temp;
	long l;
	float* x;
  
	l = 0;

	if (ReadFile(fname, f)) {
		long tl = 0;
		do {
			tl = fscanf (f, "%f", &temp);
			if (tl==1) {
				l+= tl;
			}
		} while (tl==1);
		CloseFile(f);
	} else {
		Serror ("Cannot read file %s\n",fname);
		return NULL;
	}

	if (ReadFile(fname, f)) {    
		long i;
		buffer = NewSampleBuffer(l, 0);
		if (!buffer) {
			return NULL;
		}
		x = buffer->x;

		do {
			i = fscanf(f,"%f",x++);
		} while (i==1);

		assert(l==buffer->length);
		CloseFile(f);
		return buffer;
	} 

	return NULL;
}

//==========================================================
// PrintSampleBuffer()
//----------------------------------------------------------
/*
  DESCRIPTION

  - Saves the buffer into a file, using an ASCII numerical format.

  ARGUMENTS

  - buffer: a pointer to the buffer to be saved.
  - f: A pointer to the file structure in which to save the data.

  RETURN VALUE

  - A pointer to the buffer if the operation was successful, otherwise NULL.

  SEE ALSO

  - SaveSampleBuffer(), FreeSampleBuffer()
*/
int PrintSampleBuffer(const SAMPLE_BUFFER* buffer, FILE* f) {
	register float* x;
	register long i;
	int ret;
  
	if ((ret = CheckBuffer(buffer))){
		return ret;
	}

	if (!f){
		Serror("NULL filestream\n");
		return (DEC_ARG_INVALID|2);
	}

	x = buffer->x;
	for (i=0; i<buffer->length; i++) {
		fprintf(f, "%ld %f\n", i, *x++);
	}

	return 0;
}

//==========================================================
// SaveSampleBuffer()
//----------------------------------------------------------
/*
  DESCRIPTION

  - Saves the buffer into a file, using 16-bit signed binary format.

  ARGUMENTS

  - buffer: a pointer to the buffer to be saved.
  - f: A pointer to the file structure in which to save the data.

  RETURN VALUE

  - A pointer to the buffer if the operation was successful, otherwise
  NULL.

  SEE ALSO

  - SaveSampleBuffer(), FreeSampleBuffer()
*/
int SaveSampleBuffer(const SAMPLE_BUFFER* buffer, FILE* f) {
	register float* x;
	register long i;
	short y;
	register float t;
	int ret;
  
	logmsg ("Saving buffer\n");
	if ((ret = CheckBuffer(buffer))){
		return ret;
	}

	if (!f){
		Serror("NULL filestream\n");
		return (DEC_ARG_INVALID|2);
	}

	x = buffer->x;
	for (i=0; i<buffer->length; i++) {
		t = (32767.0)*(*x++);
		y = (short) t;
		fwrite(&y, sizeof(short), 1, f);
	}

	return 0;
}




//==========================================================
// CheckBuffer()
//----------------------------------------------------------
int CheckBuffer(const SAMPLE_BUFFER* buffer) {
	if (!buffer) {
		Serror("buffer is NULL\n");
		return (DEC_ARG_INVALID|1);
	}

	if (buffer->x == NULL) {
		Serror("buffer contents NULL\n");
		return (DEC_ARG_CORRUPT|1);
	}

	if (buffer->length == 0) {
		Serror("buffer length 0\n");
		return (DEC_ARG_CORRUPT|1);
	}


	return 0;
}

//==========================================================
// FreeSampleBuffer()
//----------------------------------------------------------

int FreeSampleBuffer(SAMPLE_BUFFER* buffer) {
	if (!buffer) {
		fprintf(stderr, "FreeSampleBuffer() called with NULL argument\n");
		return (DEC_ARG_INVALID|1);
	}

	if (buffer->x) {
		free(buffer->x);
	}else{
		fprintf(stderr, "Warning: buffer contents NULL\n");
	}

	if (buffer->length == 0) {
		fprintf(stderr, "Warning: buffer length 0\n");
	}

	free(buffer);

	return 0;


}


//==========================================================
// NewBufferList
//----------------------------------------------------------
// Use this to make new, empty buffer lists.

SAMPLE_BUFFER_LIST* NewBufferList(int n_buffers)
{
	int i;
	SAMPLE_BUFFER_LIST* blist;
	blist = AllocM (SAMPLE_BUFFER_LIST, 1);
	blist->s = AllocM (SAMPLE_BUFFER*, n_buffers);

	for (i=0; i<n_buffers; i++) {
		blist->s[i] = NULL;
	} 
	blist->n = n_buffers;
	return blist;
}


//==========================================================
// LoadBufferList()
//----------------------------------------------------------
// Use that with a hook to make things easier.
SAMPLE_BUFFER_LIST* LoadBufferList(const char* const* file, int n_files, SAMPLE_BUFFER* (*load_hook) (const char* fname))
{
	int i;
	SAMPLE_BUFFER_LIST* blist;
	blist = AllocM (SAMPLE_BUFFER_LIST, 1);
	blist->s = AllocM (SAMPLE_BUFFER*, n_files);

	for (i=0; i<n_files; i++) {
		blist->s[i] = load_hook(file[i]);
		if (!blist->s[i]) {
			Serror("Did not get a sample buffer for example %d\n",i);
			return NULL;
		}
	} 
	blist->n = n_files;
	return blist;
}

//==========================================================
// LoadBufferList()
//----------------------------------------------------------

SAMPLE_BUFFER_LIST* LoadBufferList(const char* const* file, int n_files)
{
	return LoadBufferList (file, n_files, &LoadSampleBuffer);
}

//==========================================================
// SndFileBufferList()
//----------------------------------------------------------

SAMPLE_BUFFER_LIST* SndFileBufferList(const char* const* file, int n_files)
{

	return LoadBufferList (file, n_files, &SndFileSampleBuffer);

}

//==========================================================
// LoadASCIIBufferList()
//----------------------------------------------------------

SAMPLE_BUFFER_LIST* LoadASCIIBufferList(const char* const* file, int n_files)
{
	return LoadBufferList (file, n_files, &LoadASCIISampleBuffer);
}



//==========================================================
// FreeBufferList()
//----------------------------------------------------------

void FreeBufferList(SAMPLE_BUFFER_LIST* blist)
{
	int i;
	for (i=0; i<blist->n; i++) {
		FreeSampleBuffer(blist->s[i]);
	}
	FreeM(blist->s);
	FreeM(blist);
}


//==========================================================
// SplitInterleavedBuffer()
//----------------------------------------------------------
/* Splits an interleaved buffer with n_dim channels
   into an array of single-channel buffers. The initial
   buffer is not freed.
   
   RETURN VALUE
   A pointer to the array of SAMPLE_BUFFER objects, otherwise NULL
*/
SAMPLE_BUFFER_LIST* SplitInterleavedBuffer (const SAMPLE_BUFFER* buffer, int n_dim)
{
	int i, j;
	int l, q;
	SAMPLE_BUFFER_LIST* blist = NULL;
	float* data = buffer->x;
	q = (buffer->length) % n_dim;
	if (q) {
		Serror ("The number of samples in the buffer is not divisible by the number of dimensions!");
	}

	/* Allocate n_dim buffers of length l */
	l = (buffer->length) / n_dim;
	blist = AllocM (SAMPLE_BUFFER_LIST, 1);
	blist->n = n_dim;
	blist->s = AllocM (SAMPLE_BUFFER*, n_dim);
	for (i=0; i<n_dim; i++) {
		blist->s[i] = NewSampleBuffer(q, 0);
	}

	/* Copy all the data */
	for (i=0; i<n_dim; i++) {
		for (j=0; j<l; j++) {
			blist->s[i]->x[j] = *data++;
		}
	}

	return blist;
}

//==========================================================
// InterleaveBuffers
//----------------------------------------------------------
/*
  DESCRIPTION

  - Merges an array of buffers into a single interleaved buffer. The
  length of the interleaved buffer will be equal to the length of the
  largest buffer in the array. All buffers will be aligned to the
  start of the interleaved buffer and padded with zeros if they are
  shorter than its length. If one or more of the buffers are null, a
  warning is emitted, but operation continues, replacing null buffer
  contents with zeros in the interleaved stream. The original buffers
  are not freed.

  - Note that it is hard to operate on interleaved data. In most cases
  it is better to work on separate data streams. TimedStreams.c and
  TimedStreams.h provide more flexible methods for operating on
  multichannel streams.

  ARGUMENTS

  - buffer: A pointer to the buffers to be merged.
  - n_dim: The number of dimensions in the interleaved buffer.

  RETURN VALUE

  - A pointer to the interleaved buffer, or NULL if the operation was
  not successful.

  SEE ALSO

  - TimedStreams.c
*/

SAMPLE_BUFFER* InterleaveBuffers (const SAMPLE_BUFFER* const* buffer, int n_dim) {
	int i;
	long max_length = -1;
	long l;
	int max_i = 0;
	int nulls = 0;
	int no_data = 0;
	int length_mismatch = 0;
	SAMPLE_BUFFER* interleaved_buffer = NULL;

	/* Errors, mismatches and sanity checks */
	assert (buffer);

	for (i=0; i<n_dim; i++) {
		if (!buffer[i]) {
			nulls++;
		} else {
			if ((buffer[i]->length) && (buffer[i]->x)) {
				if (buffer[i]->length > max_length) {
					max_i = i;
					max_length = buffer[i]->length;
				}
			} else {
				no_data++;
			}
		}
	}

	for (i=0; i<n_dim; i++) {
		if (buffer[i]) {
			if ((buffer[i]->length) && (buffer[i]->x)) {
				if (buffer[i]->length != max_length) {
					length_mismatch++;
				}
			}
		}
	}
	if (nulls) {
		Swarning ("Found %d NULL buffers\n", nulls);
	}
	if (no_data) {
		Swarning ("Found %d empty buffers\n", no_data);
	}
	if (length_mismatch) {
		fprintf (stderr, "%d buffers are short, padding with zeros\n", length_mismatch);
	}

	if (((nulls + no_data) == n_dim)||(max_length<0)) {
		Serror ("Sorry, but none of the buffers seem to contain any data!");
	}
  
	/* Allocate buffer and clear it */
	interleaved_buffer = NewSampleBuffer (max_length*n_dim, 1);
	if (!interleaved_buffer) {
		Serror ("Failed to create interleaved buffer!\n");
		return NULL;
	}

	/* Copy data - outer loop through buffers, should be faster when
	   buffers are mismatched - otherwise swap the loops.*/
	if (nulls||no_data||length_mismatch) {
		for (i=0; i<n_dim; i++) {
			if (buffer[i]) {
				if ((buffer[i]->length)&&(buffer[i]->x)) {
					float* x = interleaved_buffer->x + i;
					float* y = buffer[i]->x;
					assert (buffer[i]->length<=max_length);
					for (l=0; l<max_length; l++) {
						*x = *y++;
						x += n_dim;
					}
				}
			}
		} 
	} else {
		float* x = interleaved_buffer->x;
		float** y; // allocate this for speed!
		y = AllocM(float*, n_dim);

		if (!y) {
			Serror ("Error allocating cache\n");
			FreeSampleBuffer(interleaved_buffer);
			return NULL;
		}
		for (i=0; i<n_dim; i++) 
			y[i] = buffer[i]->x;

		for (l=0; l<max_length; l++) {
			for (i=0; i<n_dim; i++) {
				*x++ = *(y[i]++);
			}
		}
	}   
         
	return interleaved_buffer;

}


//==========================================================
// ScaleSampleBuffer()
//----------------------------------------------------------
/*
  DESCRIPTION

  - Scales the buffer to lie within [-1.0, 1.0]

  ARGUMENTS

  - buffer

  RETURN VALUE

  - A pointer to the buffer if the operation was completed, otherwise NULL.

*/
SAMPLE_BUFFER* ScaleSampleBuffer (SAMPLE_BUFFER* buffer) {
	long l;
	float* x;
	if (CheckBuffer(buffer)) return NULL;
	float max=0.0f;

	x = buffer->x;
	for (l=0; l<buffer->length; l++) {
		float y = fabs(*x++);
		if (y>max) {
			max = y;
		}
	}
 
	assert (max>0.0f);

	max = 1.0f/max;

	x = buffer->x;
	for (l=0; l<buffer->length; l++) {
		*x++ = (*x)*max;
	}
	return buffer;
}


//==========================================================
// CopyBuffer()
//----------------------------------------------------------
/*
  DESCRIPTION

  - Creates a duplicate of the buffer

  ARGUMENTS

  - buffer.

  RETURN VALUE

  - A pointer to the newly created buffer if the operation was
  successful, othewise NULL.
*/
SAMPLE_BUFFER* CopyBuffer (const SAMPLE_BUFFER* buffer)
{
	long i;
	long l = buffer->length;
	SAMPLE_BUFFER* dst = NULL;
	float* x;
	float* y;

	//  logmsg ("Copying %ld samples\n", l);
	dst = NewSampleBuffer (l, 0);
	assert (dst);
	x = buffer->x;
	y = dst->x;
	for (i=0; i<l; i++) {
		*y++ = *x++;
	}
	return dst;
}


SAMPLE_BUFFER* CopyBufferRange (const SAMPLE_BUFFER* buffer, long start, long length){
	//  logmsg ("Copying %ld samples\n", copy_length);
	SAMPLE_BUFFER* dst = NewSampleBuffer (length, CLEAR_SAMPLE_BUFFER);
	if (CopyBufferRange (dst, buffer, start, length)) {
		Serror ("Copy Failed\n");
		FreeSampleBuffer (dst);
		return NULL;
	}
	return dst;
}


int CopyBufferRange (SAMPLE_BUFFER* dst, const SAMPLE_BUFFER* src, long start, long length)
{
	long l = src->length;
	long copy_length;
  
	NewSampleBuffer (length, CLEAR_SAMPLE_BUFFER);
	if ((dst == NULL)||(dst->x==NULL)) {
		Swarning ("NULL destination\n");
		return -1;
	}

	if (start<0) {
		Swarning ("Negative index\n");
		return -1;
	}

	if (l<(start+length)) {
		Swarning ("Source (%ld < %ld+%ld = %ld)copy too large, padding with 0s\n", l, start, length, start+length);
		copy_length = l - start;
	} else {
		copy_length = length;
	}
	if (copy_length > dst->length) {
		Swarning ("Destination too small for copy, truncating");
		copy_length = dst->length;
	}

	register float* x = src->x;
	register float* y = dst->x;
	for (register int i=0; i<copy_length; i++) {
		*y++ = *x++;
	}
	dst->rate = src->rate;
	return 0; 
}


//==========================================================
// ExtendBuffer()
//----------------------------------------------------------
/*
  DESCRIPTION

  - Creates a duplicate of the buffer and frees the first

  ARGUMENTS

  - buffer.

  RETURN VALUE

  - A pointer to the newly created buffer if the operation was
  successful, othewise NULL.
*/
SAMPLE_BUFFER* ExtendBuffer (SAMPLE_BUFFER* buffer, long length)
{
	long i;
	long l = buffer->length;
	SAMPLE_BUFFER* dst = NULL;
	float* x;
	float* y;

	//  logmsg ("Copying %ld samples\n", l);
	dst = NewSampleBuffer (l+length, CLEAR_SAMPLE_BUFFER);
	assert (dst);
	x = buffer->x;
	y = dst->x;
	for (i=0; i<l; i++) {
		*y++ = *x++;
	}
	FreeSampleBuffer (buffer);
	return dst;
}

//==========================================================
// MixBuffers()
//----------------------------------------------------------
/*
  DESCRIPTION

  - Mixes a number of buffers, scaling the result to the range [-1.0, 1.0]

  ARGUMENTS

  - n: The number of buffers to be mixed.
  - src: An array of buffer pointers from which the data is to be taken.
  - dst: The destination buffer. If NULL, a new one is created.

  RETURN VALUE

  - NULL if the operation fails, otherwise a pointer to the mixed buffer.

*/
SAMPLE_BUFFER* MixBuffers (int n, const SAMPLE_BUFFER* const* src, SAMPLE_BUFFER* dst)
{
	int i;
	long length = 0;

	logmsg ("mixing...\n");
	for (i=0; i<n; i++) {
		if (src[i]==NULL) {
			Serror ("Buffer %d is null! Aborting mix\n", i);
			return NULL;
		}
		if ((src[i]->length) >= length) {
			length = src[i]->length;
		}
	}

	if (dst==NULL) {
		fprintf (stderr, "No destination buffer supplied, creating new one with length %ld\n", length);
		dst = NewSampleBuffer (length, CLEAR_SAMPLE_BUFFER);
	}
  

	assert (dst);
	assert (dst->length >= length);
	assert (dst->x);

	for (i=0; i<n; i++) {
		if (src[i]==dst) {
			Swarning ("Warning: source buffer %d same as destination\n", i);
		}
	}

	for (i=0; i<n; i++) {
		int l;
		printf ("..buffer %d/%d..\n",1+i,n);

		if ((src[i]->length==0)||(src[i]->x==NULL)) {
			Swarning ("Buffer %i seems to contain no data. Ignoring.\n", i);
		} else {
			float* y = dst->x;
			float* x = src[i]->x;
			assert (src[i]->length <= length);
			for (l=0; l<src[i]->length; l++) {      
				*y++ += *x++;
			}
		}
	}
	ScaleSampleBuffer (dst);
	return dst;
}

SAMPLE_BUFFER* GetListSample (const SAMPLE_BUFFER_LIST* buffer_list, int i)
{
	if ((i<0) || (i>=buffer_list->n)) {
		Serror ("Index out of range\n");
	}
	assert (buffer_list->s);
	assert (buffer_list->s[i]);
	assert (buffer_list->s[i]->x);
	return (buffer_list->s[i]);
}
