/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* "VER: $Id: SampleBuffer.h,v 1.20 2003/09/06 17:26:18 olethros Exp olethros $" */
/* SampleBuffer.h : Create, read and write sample buffers */

/*! \file SampleBuffer.h
 * \brief Low-level handling of sample buffers.
 *
 * The buffers just store data, without knowing what format they are
 * in. For example in most cases multiple channel data is stored in an
 * interleaved format. Sometimes though the data might be concatenated
 * or stored in different files. SampleBuffer does not really care
 * about that.  Methods to deal with different data formats should be
 * provided in support code for the particular data format you are
 * interested in. In other words, the client code is responsible for
 * keeping track of the data tyep in each SAMPLE_BUFFER they use.
 * 
 * - When doing batch processing across multiple channels, it is
 * useful to store the data in an interleaved format. For other
 * applications, such as multi-track music mixing it is more useful to
 * store each channel in a separate buffer and use some other
 * structure to hold the data in.
 *
 * - A method for splitting an interleaved buffer into separate buffers
 * is provided, called SplitInterleavedBuffer()
 *
 * - The reverse method is provided, but the interleaving of streams
 * with unequal length should depend upon the application. It is suggested
 * to use the method in TimedStreams instead.
 *
 */

#ifndef SAMPLEBUFFER_H
#define SAMPLEBUFFER_H

#include <stdio.h>
#include <stdlib.h>
#include <sndfile.h>

/// Sample buffer implementation.
typedef struct SampleBuffer{
  long length; ///< The buffer length
  float *x; ///< Pointer to buffer contents
  int rate; ///< Sampling rate
} SAMPLE_BUFFER;

/// Sample buffer list
/*
  A SamleBufferList implements a sized array of sample buffers. 
 */
typedef struct SampleBufferList {
  int n;
  SAMPLE_BUFFER** s;
} SAMPLE_BUFFER_LIST;

/// Create a new sample buffer
SAMPLE_BUFFER* NewSampleBuffer (long length=0, int flags=0);

/// Free a sample buffer
int FreeSampleBuffer (SAMPLE_BUFFER* buffer);

/// Clear the buffer contents
void ClearBuffer (SAMPLE_BUFFER* buffer);

/// Set the buffer length to a particular value
long SetBufferLength (SAMPLE_BUFFER* buffer, long length);

/// Request a minimum buffer length
long RequestBufferLength (SAMPLE_BUFFER* buffer, long length);

/// Return a copy of a buffer range
SAMPLE_BUFFER* CopyBufferRange (const SAMPLE_BUFFER* buffer, long start, long length);
/// Copy a buffer range to another buffer
int CopyBufferRange (SAMPLE_BUFFER* dst, const SAMPLE_BUFFER* src, long start, long length);
/// Return a copy of the buffer
SAMPLE_BUFFER* CopyBuffer (const SAMPLE_BUFFER* buffer);

/// Add some length to a buffer. Result in a new buffer
SAMPLE_BUFFER* ExtendBuffer (SAMPLE_BUFFER* buffer, long length);

/// Mix a number of buffers to a destination buffers
SAMPLE_BUFFER* MixBuffers (int n, const SAMPLE_BUFFER* const* src, SAMPLE_BUFFER* dst);

/// Sanity check
int CheckBuffer (const SAMPLE_BUFFER* buffer);

/// Save a buffer as ASCII
int PrintSampleBuffer (const SAMPLE_BUFFER* buffer, FILE* f);

/// Save a buffer in binary format
int SaveSampleBuffer (const SAMPLE_BUFFER* buffer, FILE* f);

/// Load raw data
SAMPLE_BUFFER* LoadSampleBuffer (const char* fname);

/// Load data with libsndfile
SAMPLE_BUFFER* SndFileSampleBuffer (const char* fname);

/// Load raw ASCII data
SAMPLE_BUFFER* LoadASCIISampleBuffer (const char* fname);

/// Make a buffer list
SAMPLE_BUFFER_LIST* NewBufferList (int n_buffers);

/// Make a buffer list, loading sound data from files
SAMPLE_BUFFER_LIST* SndFileBufferList (const char* const* file, int n_files);

/// Make a buffer list, loading raw data from files
SAMPLE_BUFFER_LIST* LoadBufferList (const char* const* file, int _files);

/// Make a buffer list, loading ASCII data from files
SAMPLE_BUFFER_LIST* LoadASCIIBufferList (const char* const* file, int n_files);

/// Free a buffer list
void FreeBufferList (SAMPLE_BUFFER_LIST* buffer);

/// Split an interleaved buffer into a list
SAMPLE_BUFFER_LIST* SplitInterleavedBuffer (const SAMPLE_BUFFER* buffer, int n_dim);

/// Combine buffers into an interleaved buffer
SAMPLE_BUFFER* InterleaveBuffers (const SAMPLE_BUFFER* const * buffer, int n_dim);

/// Scale a sample buffer to 1.0
SAMPLE_BUFFER* ScaleSampleBuffer (SAMPLE_BUFFER* buffer);

/// Get i'th sample buffer from a list
SAMPLE_BUFFER* GetListSample (const SAMPLE_BUFFER_LIST* buffer_list, int i);

/// Flags for file formats
#define UNKNOWN_FILETYPE 0x0
#define FLOAT_FILETYPE 0x1
#define UNSIGNED_FILETYPE 0x02
#define COMPLEMENT_FILETYPE 0x04
#define BIT8_FILETYPE 0x8
#define BIT16_FILETYPE 0x10
#define BIT32_FILETYPE 0x20
#define TYPE_LITTLE_ENDIAN 0x40
#define BITMASK_FILETYPE 0x38

/// Flags for quick operations, as accepted by functions accepting flags*/
#define CLEAR_SAMPLE_BUFFER 0x1

#endif
