/* -*- Mode: C++ -*- */
/* $Id: SegmentedSample.c,v 1.6 2003/09/02 19:24:49 olethros Exp $" */
#include "SegmentedSample.h"


/* Given segment labels and lists of inputs and targets, we generate one big list of input/targets with segmentation*/
IOExampleList* LabelSegments (FilePair* pairs, LINKEDLIST* labels, int cutoff)
{
  IOExampleList* xl = NULL;
  char** input_files = pairs->input_files;
  char** target_files = pairs->target_files; 
  long n_files = pairs->n_files;
  long n_dim  = pairs->n_dim;
  
  if (n_dim != 2) {
    Serror ("Expected pairs, found %d columns", n_dim);
    return NULL;
  }

  if ((input_files == NULL) || (target_files == NULL)) {
    Serror ("Stupid pointers\n");
    return NULL;
  }

  if ((xl = AllocM (IOExampleList, 1)) == NULL) {
    Serror ("Could not allocate list\n");
    return NULL;
  }

  if ((xl->ex_list = List ()) == NULL) {
    Serror ("List creation failed\n");
    FreeM (xl);
    return NULL;
  }

  message ("Loading files - adding %d samples to segments", cutoff);
  xl->n  = 0;
  for (int i=0; i<n_files; i++) {
    SegmentList* seg;
    SAMPLE_BUFFER_LIST* buffers;
    
    assert (input_files[i]);
    assert (target_files[i]);
    
    seg = LoadHTKSegmentation (target_files[i], labels);
    buffers = LoadSampleSegments (input_files[i], seg, TIME_UNITS_MS, cutoff);
    AddExamplesToList (xl, buffers, seg);
  }

  message ("Added %d files (%d examples)", n_files, xl->n);
  return xl;
}

void FreeIOExampleList (IOExampleList* xl)
{
  ClearList (xl->ex_list);
  FreeM (xl);
}

int SelectFirstExample (IOExampleList* xl)
{
  LISTITEM* li;
  li = FirstListItem (xl->ex_list);
  xl->index = 0;
  xl->current = (IOExample*) li->obj;
  GetIndexedExample (xl);
  return 0;
}

int SelectNextExample (IOExampleList* xl)
{
  LISTITEM* li;
  IOExample* curr = xl->current;
  xl->index++;
  if (xl->index >= curr->samples->n) {
    xl->index = 0;
    li = NextListItem (xl->ex_list);
    if (li == NULL) {
      xl->current = NULL;
      return 1;
    }
    xl->current = (IOExample*) li->obj;
  }  
  GetIndexedExample (xl);
  return 0;
}

void GetIndexedExample (IOExampleList* xl)
{
  IOExample* curr = xl->current;
  int i = xl->index;

  xl->input = GetListSample (curr->samples, i);
  xl->target = curr->target[i];
}

int AddExamplesToList (IOExampleList* xl, SAMPLE_BUFFER_LIST* buffers, SegmentList* seg)
{
  IOExample* example;
  int n_examples;
  
  if (buffers == NULL) {
    Serror ("NULL buffers\n");
    return DEC_ARG_INVALID;
  }

  if ((example = AllocM (IOExample, 1)) == NULL) {
    Serror ("Memory error\n");
    return DEC_MEM_ERROR;
  }

  if ((n_examples = buffers->n) <= 0) {
    Serror ("Sanity Error\n");
    return DEC_SANITY_ERROR;
  }

  xl->n += buffers->n;
  example->samples = buffers;
  example->target = GetAllSegmentTargets (seg);
  FreeSegmentList (seg); // we do not need that any more.
  ListAppend (xl->ex_list, example, &FreeIOExample);
  return 0;
}

void FreeIOExample (void* obj) 
{
  IOExample* example = (IOExample*) obj;
  
  FreeM (example->target);
  FreeBufferList (example->samples);
  FreeM (example);
  
}

 

int SegmentStart (SegmentList* seglist, int i)
{
     LISTITEM* li;
     Segment* s;
     if ((li = GetItem (seglist->list, i)) == NULL) {
	  return -1;
     }
     s = (Segment*) li->obj;
     return s->start_time;

}
int SegmentEnd (SegmentList* seglist, int i)
{
     LISTITEM* li;
     Segment* s;
     if ((li = GetItem (seglist->list, i)) == NULL) {
	  return -1;
     }
     s = (Segment*) li->obj;
     return s->end_time;
}

char* SegmentLabel (SegmentList* seglist, int i)
{
     LISTITEM* li;
     Segment* s;
     if ((li = GetItem (seglist->list, i)) == NULL) {
       return NULL;
     }
     s = (Segment*) li->obj;
     return s->label_name;
}


SAMPLE_BUFFER_LIST* LoadSampleSegments (char* sample_fname, SegmentList* seglist, TimeStamp time_units, int play_room)
{
    SNDFILE* tmp = NULL;
    SF_INFO file_info;
    SAMPLE_BUFFER_LIST* buffer_list = NULL;
    SAMPLE_BUFFER* tmp_buffer = NULL;
    int rate = 0;

    /* Open the sample */
    tmp = sf_open (sample_fname, SFM_READ, &file_info);
    if (tmp) {
      if (!(sf_format_check(&file_info))) {
        Serror ("Invalid format for %s?\n", sample_fname);
        sf_close (tmp);
        return NULL;
      }
    } else {
      Serror ("Could not open %s\n", sample_fname);
    }


    rate = file_info.samplerate;
    
    tmp_buffer = NewSampleBuffer (file_info.frames + 64, CLEAR_SAMPLE_BUFFER);
    sf_read_float (tmp, tmp_buffer->x, file_info.frames);

    buffer_list = NewBufferList (n_Segments (seglist));
    TimeStamp total_length = 0;
    TimeStamp sample_length = file_info.frames;

    for (int i=0; i<n_Segments (seglist); i++) {
        TimeStamp start_time, end_time, length;
        int n_samples;

        start_time = SegmentStart (seglist, i);
        end_time = SegmentEnd (seglist, i);
        length = end_time - start_time;
        assert (length > 0);
        n_samples = length*rate/time_units;
        total_length += n_samples;

        if (n_samples<play_room*2) {
          message ("N:%d L:%s", n_samples, SegmentLabel (seglist, i));
        }

        if (total_length>sample_length) {
          Swarning ("Segment %d/%d: total length seems to be larger than the length of the sample. Leaving sample buffer empty\n", i, n_Segments (seglist));
          buffer_list->s[i] = NewSampleBuffer (play_room*2, CLEAR_SAMPLE_BUFFER);
          buffer_list->s[i]->rate = rate;
        } else {
          TimeStamp start_sample=start_time*rate/time_units;
          if (start_sample > play_room) {
            buffer_list->s[i] = CopyBufferRange (tmp_buffer, start_sample-play_room, n_samples+play_room);
          } else {
            buffer_list->s[i] = CopyBufferRange (tmp_buffer, start_sample, n_samples+play_room);
          }
          buffer_list->s[i]->rate = rate;
        }
    }

    FreeSampleBuffer (tmp_buffer);
    sf_close(tmp);

    
    return buffer_list;
}

//====================================================================
// loadHTKSegmentation()
//--------------------------------------------------------------------
/* ARGUMENTS
   - fname: the name of the file to be scanned.
   - labes: A linkedlist of labels.

   RETURNS
   - A pointer to a new SegmentList

   DESCRIPTION

   This function reads a file of type
   start0 end0 label0
   start1 end1 label0
   ...

   Where startN, endN are start and end times for the part of the
   sequence labelled with labelN. It assumes that the times are in
   nanoseconds and stores the values in the SegmentList structure,
   along with the labels.

   In order to assign unique numeric identifiers to labels, it uses
   the labels LINKEDLIST. If labels is NULL, all identifiers are set
   to -1, however the label string is stored.
*/
SegmentList* LoadHTKSegmentation(char* fname, LINKEDLIST* labels)
{
    FILE* fs;
    int start_time;
    int end_time;
    StringBuffer* sb = NULL;
    SegmentList* seglist = NULL;

    if ((fs = fopen (fname, "r")) == NULL) {
	 return NULL;
    }
    
    seglist = NewSegmentList();

    do {
	 int i;
	 sb = read_string (fs, sb);
	 char* s;
	 char* label = NULL;
	 int label_id;
	 if (sb->string) {
	      s = strtok (sb->string, ", \n");
	      i = 0;
	      while (s) {
		   switch (i) {
		   case 0: sscanf(s, "%d", &start_time);break;
		   case 1: sscanf(s, "%d", &end_time);break;
		   case 2: label = string_copy (s);
		   }
		   i++;
		   s = strtok(NULL, ", \n");
	      }
	      if (labels) {
		   label_id = (GetStringMatchingListItem (labels, label))->ID;
		   AddSegmentToList (seglist, NewSegment((TimeStamp) start_time, 
                                             (TimeStamp) end_time,
                                             label, label_id));
	      } else {
		   AddSegmentToList (seglist, NewSegment((TimeStamp) start_time,
                                             (TimeStamp) end_time,
                                             label, -1));
	      }
	 }
    } while (sb->string);

    fclose (fs);
    FreeStringBuffer (&sb);
    return seglist;
} 

LINKEDLIST* LoadHTKLabels (char* fname) 
{
     LINKEDLIST* labels;
     StringBuffer* sb = NULL;

     FILE* fs;
     if (fname == NULL) {
       Serror ("Expected a filename\n");
       return NULL;
     }
     if ((fs = fopen (fname, "r")) == NULL) {
	  Serror ("Could not open file\n");
	  return NULL;
     }

     if ((labels = NewList()) == NULL) {
	  Serror ("Could not make label list\n");
	  fclose (fs);
	  return NULL;
     }

     sb = read_string (fs, sb);
     if (sb->string == NULL) {
	  Serror ("File is empty\n");
     }

     while (sb->string) {
	  char* s = strtok (sb->string, ", \n");
	  GetStringMatchingListItem (labels, s);
	  sb = read_string (fs, sb);
     }
     fclose (fs);
     FreeStringBuffer (&sb);
     return labels;
}

int n_Segments (SegmentList* seglist)
{
    return ListSize (seglist->list);
}

SegmentList* NewSegmentList(void)
{
     SegmentList* sl = NULL;
     sl = AllocM (SegmentList, 1);
     sl->list = List ();
     return sl;
}

void FreeSegmentList (SegmentList* seglist)
{
     ClearList (seglist->list);
     FreeM (seglist);
}

Segment* NewSegment (TimeStamp start_time, TimeStamp end_time, char* label, int label_id)
{
     Segment* seg = NULL;
     seg = AllocM (Segment, 1);
     seg->start_time = start_time;
     seg->end_time = end_time;
     seg->label_name = label;
     seg->label_id = label_id;
     return seg;
}


void FreeSegment (void* segment)
{
     Segment* s = (Segment*) segment;
     FreeM (s->label_name);
     FreeM (s);
}

void AddSegmentToList (SegmentList* seglist, Segment* segment)
{
     ListAppend (seglist->list, segment, &FreeSegment);
}

void ShowSegmentation (SegmentList* seg)
{
     LISTITEM* li;

     li = FirstListItem (seg->list);
     while (li) {
	  Segment* s = (Segment*) li->obj;
	  printf ("%d %d %s %d\n", s->start_time, s->end_time, s->label_name, s->label_id);
	  li = NextListItem (seg->list);
     }
}


int* GetAllSegmentTargets (SegmentList* seg)
{
  LISTITEM* li;
  int* t;
  int n_seg = n_Segments (seg);
  int i = 0;
  t = AllocM (int, n_seg);
  
  li = FirstListItem (seg->list);
  while (li) {
	  Segment* s = (Segment*) li->obj;
    t[i] = s->label_id;
    li = NextListItem (seg->list);
    i++;
  }
  if (i!=n_seg) {
    Serror("Sanity error\n");
    FreeM (t);
    return NULL;
  }

  return t;
}
