/* -*- Mode: C++ -*- */
/* $Id: SegmentedSample.h,v 1.6 2003/09/03 11:36:50 olethros Exp $
 *
 * AUTHORS AND COPYRIGHT
 *
 * Copyright (C) 2003, Christos Dimitrakakis
 * 
 * This code can be freely distributed under the GNU General Public
 * License. See COPYING.
 * 
 * DESCRIPTION
 *
 * These routines can be used to split up a continous sample into a
 * number of segments, given either an already loaded sample or a
 * filename from which the sample data will be read. The segmentation
 * information can be read from another file, which contains timing
 * and labelling information for segments. 
 *
 * USAGE
 *
 * Call LoadHTKLabels() to load a list of allowed
 * labels from a file, or create an empty LINKEDLIST using
 * NewList(). Pass a pointer to the label list, or NULL, to
 * LoadHTKSegmentation, along with a file describing the segmentation
 * of your sample. The returned value can then be passed to
 * LoadSampleSegments() in order to directly load sample segments into
 * memory, or to SegmentSample(), which will segment an existing
 * sample.
 *
 */
#ifndef SEGMENTED_SAMPLE_H
#define SEGMENTED_SAMPLE_H

#include "SampleBuffer.h"
#include "List.h"
#include "LinkedList.h"
#include "debug.h"
#include "datafile.h"
#include "string_utils.h"
#include "TimeStamp.h"




typedef struct IOExample_ {
  SAMPLE_BUFFER_LIST* samples;
  int* target;
} IOExample;

typedef struct IOExampleList_ {
  LIST* ex_list;
  int index;
  IOExample* current;
  SAMPLE_BUFFER* input; //current input
  int target; //current target
  int n; // Number of examples
} IOExampleList;

typedef struct Segment_ {
    TimeStamp start_time;  //segment start time
    TimeStamp end_time;    //segment end time
    char* label_name;  //segment label name
    int label_id;      //segment label ID
} Segment;

typedef struct SegmentList_ {
    LIST* list;
}SegmentList;


/* Highest level functions */
/* Given segment labels and lists of inputs and targets, we generate one big list of input/targets with segmentation*/
IOExampleList* LabelSegments (FilePair* pairs, LINKEDLIST* labels, int cutoff = 0);
void FreeIOExampleList (IOExampleList* xl);

int SelectFirstExample (IOExampleList* xl);
int SelectNextExample (IOExampleList* xl);
void GetIndexedExample (IOExampleList* xl);

/* Middle-level functions */

/* Given a sample name and a list of segments, create an array of
 * buffers corresponding to the sample segments*/
SAMPLE_BUFFER_LIST* LoadSampleSegments (char* sample_fname, SegmentList* seglist, TimeStamp time_units, int play_room = 16);
/* Create a segment list from an HTK Segmentation */
SegmentList* LoadHTKSegmentation(char* fname, LINKEDLIST* labels);
/* Load labels from this filename */
LINKEDLIST* LoadHTKLabels (char* fname);
/* The number of segments in the segment list */
int n_Segments (SegmentList* seglist);
/* Shows timings and labels for debugging purposes */
void ShowSegmentation (SegmentList* seg);
int* GetAllSegmentTargets (SegmentList* seg);

/* Low-level functions */
SegmentList* NewSegmentList(void);
void FreeSegmentList (SegmentList* seglist);
int SegmentStart (SegmentList* seglist, int i);
int SegmentEnd (SegmentList* seglist, int i);
char* SegmentLabel (SegmentList* seglist, int i);
Segment* NewSegment (TimeStamp start_time, TimeStamp end_time, char* label, int label_id);
void FreeSegment (void* seg);
void AddSegmentToList (SegmentList* seglist, Segment* segment);

int AddExamplesToList (IOExampleList* xl, SAMPLE_BUFFER_LIST* buffers, SegmentList* seg);
void FreeIOExample (void* obj) ;

#endif
