#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include "real.h"
#include "debug.h"


/* This program is used to analyse the performance of boosting when
 * the roles are inversed.  Specifically, we are looking for what happens
 * when we combine inaccurate classifiers.
 * 
 */

typedef struct Classifier_ {
  real* likelihood;
  real accuracy;
} Classifier;

void DoSoftmax(real* x, int l);
void Simulate (int n_boost, int n_classes);
int get_max_id (real* y, int l);

int main (void) {

  /* Parameters */
  int n_boost = 3;
  int n_classes = 10;


  srand48(time(NULL));

  for (n_boost=1; n_boost<256; n_boost++) {
    Simulate (n_boost, n_classes);
  }

  return 0;
}

void Simulate (int n_boost, int n_classes)
{
  /* Dynamically allocated */
  Classifier* model;
  real* w;
  real* y;
  real* hmm;
  
  /* scratch */
  int i,j;
  real r = 0;
  int iteration;
  int n;
  int n_iterations = 10000;
  int boost_correct = 0;
  int boostb_correct = 0;
  int weight_correct = 0;
  int weightb_correct = 0;

  /* setup */

  model = AllocM (Classifier, n_boost);
  w = AllocM (real, n_boost);
  y = AllocM (real, n_classes);
  hmm = AllocM (real, n_classes*n_boost);
  assert (model);
  assert (w);
  assert (y);
  assert (hmm);

  for (i=0; i<n_boost; i++) {
    real p = 1.0f;
    model[i].likelihood = AllocM (real, n_classes);
    assert (model[i].likelihood);
    w[i] = 1-((real) i)/((real)(n_boost));
    for (j=0; j<n_classes; j++) {
      model[i].likelihood[j] = p + (1.0f-w[i])*drand48();
      p = p*0.1;
    }
    DoSoftmax (model[i].likelihood, n_classes);

  }
  

  DoSoftmax (w, n_boost);

  /* Show what happens in adaboost */
  for (iteration=0; iteration<n_iterations; iteration++) {

    /* ADABOOST */
    for (i=0; i<n_classes; i++) {
      y[i] = 0.0f;
    }
    for (i=0; i<n_boost; i++) {
      real sum = 0.0f;
      r = drand48();
      for (j=0; j<n_classes; j++) {
	sum += model[i].likelihood[j];
	if (r<sum) {
	  y[j] += w[i];
	  break;
	}
      }
    }
    if (get_max_id (y, n_classes) == 0) {
      boost_correct++;
    }

    /* ADABOOST 2*/
    for (i=0; i<n_classes; i++) {
      y[i] = 0.0f;
    }
    for (i=0; i<n_boost; i++) {
      real sum = 0.0f;
      for (j=0; j<n_classes; j++) {
	r = drand48();
	sum = model[i].likelihood[j];
	if (r<sum) {
	  y[j] += w[i];
	}
      }
    }
    if (get_max_id (y, n_classes) == 0) {
      boostb_correct++;
    }



    /* WEIGHTED LIKELIHOODS */
    for (i=0; i<n_classes; i++) {
      y[i] = 0.0f;
    }
    for (i=0; i<n_boost; i++) {
      real sum = 0.0f;
      for (j=0; j<n_classes; j++) {
	r = drand48();
	sum = w[i]*model[i].likelihood[j];
	if (r<sum) {
	  y[j] += 1.0f;//model[i].likelihood[j];
	}
      }
    }
    if (get_max_id (y, n_classes) == 0) {
      weight_correct++;
    }


    /* WEIGHTED LIKELIHOODS 2 */
    n = 0;
    for (i=0; i<n_classes; i++) {
      y[i] = 0.0f;
      for (j=0; j<n_boost; j++, n++) {
	hmm[n] = w[j]*model[j].likelihood[i];
      }
    }

    n = 0;
    for (i=0; i<n_classes; i++) {
      real sum = 0;
      r = drand48();
      for (j=0; j<n_boost; j++, n++) {
	sum += hmm[n];
      }
      if (r<sum) {
	y[i] += 1.0f;
	break;
      }
    }


    
    if (get_max_id (y, n_classes) == 0) {
      weightb_correct++;
    }
    
      
  } /* for (iteration) */


  printf ("%f ", (((real) boost_correct)/ ((real)n_iterations)));
  printf ("%f ", (((real) boostb_correct)/ ((real)n_iterations)));

  printf ("%f ", (((real) weight_correct)/ ((real)n_iterations)));
  printf ("%f\n", (((real) weightb_correct)/ ((real)n_iterations)));

  /* Free */
  FreeM (hmm);
  for (i=0; i<n_boost; i++) {
    FreeM (model[i].likelihood);
  }
  FreeM (model);
  FreeM (w); 
  FreeM (y);
  
}



void DoSoftmax(real* x, int l)
{
  int i;
  real* e = x;
  real sum = 0.0f;
  assert (e);
  for (i=0; i<l; i++) {
    e[i] = exp(x[i]);
    sum += e[i];
  }
  assert (sum>0);
  for (i=0; i<l; i++) {
    e[i] = e[i]/sum;
  }
}

int get_max_id (real* y, int l) {
  int maxi=0;
  real max = y[maxi];
  int i;

  for (i=0; i<l; i++) {
    if (y[i]>max) {
      max = y[i];
      maxi = i;
    }
  }
  return maxi;
}
