/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
const char* CROSSBINSEQ_VERSION =  "VER: $Id: crossbinseq.c,v 1.4 2003/08/28 00:08:28 olethros Exp $";
const char* VERSION_INFO = "\n\
\tCross-combine binary sequences version 0.1\n\n\
\tCopyright (C) 2002 Dimitrakakis Christos\n\n\
This program comes comes with no WARRANTY, to the extent permitted by law.\n\
You may redistribute copies of it under the terms of the GNU General Public\n\
License. For more information about these matters see the files named COPYING.\
\n\n";

const char* help = "\n\
Cross-Combine Binary Sequences version 0.1\n\n\
Usage: crossbinseq [options]\n\n\
Cross-combine binary sequences.\n\n\
  crossbinseq <options>\n\n\
      --adapt               Use an adaptive delay for aligning, which is\n\
                            most useful when there is high jitter.\n\
      --frequency=N         set the sequence generator frequency to N. Note\n\
                            that different sequence generators employ this\n\
                            value differently\n\
      --length=N            set the generated sequence length to N.\n\
      --sine                use a sine wave as the original sequence. The\n\
                            --frequency parameter specifies its frequency\n\
      --chaotic2            use a 2-dimensional chaotic sequence as input\n\
      --chaotic1            use a 1-dimensional chaotic sequence as input\n\
      --psr                 use a pseudorandom binary sequence as the\n\
                            original sequence. The probability of a 1 equals\n\
                            what was specified in --frequency\n\
                            average \n\
      --filename            use a file as input. See also --dimensions\n\
      --dimensions=N        the number of dimensions in the input file\n\
      --ascii               use ascii for input and output\n\
      --bin                 16-bit signed mode for input and output\n\
      --models=N            specify the number of models that will be\n\
                            simulated.\n\
      --variance=N          specify how much the models will vary from the\n\
                            true sequence in amplitude\n\
      --jitter=N            amount of jitter (defaults to 1.0)\n\
      --log=filename        writes the detailed log output here\n\
      --output=filename     the name of the results file\n\
  -h, --help                display this help and exit\n\
  -v, --version             output version information and exit\n\n\
Report bugs to Christos Dimitrakakis <dimitrak@idiap.ch>\n\n";


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <getopt.h>
#include <time.h>
#include "SampleBuffer.h"
#include "psr.h"
#include "debug.h"
#include "real.h"

#define ABORT_MAIN {exit_condition = EXIT_FAILURE; goto quick_exit;}

enum SequenceType {SeqUnspecified, SeqUnknown, SeqPSR, SeqSine, SeqChaos1, SeqChaos2, SeqFromFile};

/* The PredictorModel attempts to simulate the imprecise and irregular
behaviour of actual predictors. It is hoped that by quantifying their
irregularities it will be possible to make a good estimate of
how they perform. */
typedef struct PredictorModel_ {
  long position; // position in the input stream
  real jitter; // amount of jitter in the predictions
  real variance; // predictor variance
  real bias; // predictor bias
  /* The two values below are calculated from the bias and variance */
  real fpe; // probability of a false positive
  real fne; // probability of a false negative
  real output; // this is just the model's output.
} PredictorModel;

// Create a new predictor model
PredictorModel* NewPredictorModel (real jitter, real variance, real bias);

// Delete a previously allocated model. See below.
void FreePredictorModel (PredictorModel* model);

// Use this macro instead of FreePredictorModel, since it is safer.
#define DeletePredictorModel(x) if (x) {FreePredictorModel (x); x = NULL;} else fprintf (stderr, "predictor model pointer null\n");

// Get the output
real GetOutput (PredictorModel* model, SAMPLE_BUFFER* buffer);


//=========================================================================//
int main(int argc, char* argv[])                          
{

  /* Scratch variables */
  int i;
  long t;

  /* Control variables */
  int exit_condition = EXIT_SUCCESS;   /* should be zero if all is OK */

  /* Options and config variables */
  char* output_name = NULL;
  SAMPLE_BUFFER* input_sequence = NULL;
  real frequency = 0.1f;
  enum SequenceType sequence_type = SeqUnspecified;

  /* Optional options */
  long length = 4096; // length of the input sequence
  int n_models = 8; // number of models to be simulated
  long seed = 0;
  real jitter = 1.0f;
  real variance = 0.01f;
  real bias = 0.1f;
  bool adapt = false;

  /* Files */
  FILE* output_stream = NULL;
  FILE* msgport = stderr; // <<--- does not need to be closed always.

  /* Dynamically allocated */
  PredictorModel** model = NULL;
  real* y = NULL;

  //------------------------- OPTION PROCESSING ---------------------------//
  while (1) {
    int c;
    int option_index = 0;
    static struct option long_options[] = {
      {"help", no_argument, 0, 'h'}, //0
      {"version", no_argument, 0, 'v'}, //1
      {"output", required_argument, 0, 0}, //2
      {"log", required_argument, 0, 0}, //3
      {"frequency", required_argument, 0, 0},//4
      {"sine", no_argument, 0, 0},//5
      {"psr", no_argument, 0, 0},//6
      {"models", required_argument, 0, 0},//7
      {"variance", required_argument, 0, 0},//8
      {"jitter", required_argument, 0, 0},//9
      {"length", required_argument, 0, 0},//10
      {"adapt", no_argument, 0, 0},//11
      {0, 0, 0, 0}
    };

    c = getopt_long (argc, argv, "vh",
                     long_options, &option_index);
    if (c == -1) {
      break;
    }
    switch (c) {
    case 0: /* This is the case for long arguments */
      switch(option_index) {
      case 2: output_name = optarg; break;
      case 3: Swarning ("Log file option disabled\n"); break;
      case 4: frequency = atof (optarg); break;
      case 5: sequence_type = SeqSine; break;
      case 6: sequence_type = SeqPSR; break;
      case 7: n_models = atoi (optarg); break;
      case 8: variance = atof (optarg); break;
      case 9: jitter = atof (optarg); break;
      case 10: length = atoi (optarg); break;
      case 11: adapt = true; break;
      default :
        Swarning ("Option no %d not implemented. Sorry!\n",option_index);
      }
      break;
      
    case 'h':
      printf("%s",help);
      exit(0);
      break;
      
    case 'v':
      printf("%s",VERSION_INFO);
      exit(0);
      break;
      
    case '?':
      exit (-1);

    case ':':
      exit (-1);

    default:
      fprintf (stderr, "?? getopt returned character code 0%o ??\n", c);
      exit (-1);
    }
  }
   
  if (optind < argc) {
    fprintf (stderr, "Stray arguments : ");
    while (optind < argc)
      fprintf (stderr, "%s ", argv[optind++]);
    fprintf (stderr, "\n");
    exit(-1);
  }

  //--------------------------------- SETUP --------------------------------//

  printf("%s\n",CROSSBINSEQ_VERSION);
  if (!seed) {
    seed = time(NULL);
  }
  printf("\nSeed: %ld\n\n",seed);
  srand48(seed);

  if (!output_name) {
    Serror ("No output name specified.\n");
    ABORT_MAIN;
  }

  if (!(WriteFile (output_name, output_stream))) {
    Serror ("Could not write to the file!\n");
    ABORT_MAIN;
  }
  
  switch (sequence_type) {
  case SeqUnspecified:
    Serror ("You have not specified the type of input sequence.\n");
    ABORT_MAIN;
    break;
  case SeqUnknown:
    Serror ("You seem to have specified an unknown type of sequence, perhaps because of a peculiar combination of parameters. This was unexpected.\n");
    ABORT_MAIN;
    break;
  case SeqPSR:
    input_sequence = CreatePseudoRandomSequence (length, frequency);
    break;
  case SeqSine:
    input_sequence = CreateSine (length, 44100, frequency, 0.5f);
    for (t=0; t<input_sequence->length; t++) {
      input_sequence->x[t] = input_sequence->x[t] + 0.5f;
    }
    break;
  default:
    Serror ("This sequence type is not currenlty available\n");
    ABORT_MAIN;
  }
  if (!input_sequence) {
    Serror ("Could not allocate the sequence!\n");
    ABORT_MAIN;
  }

  //-------------------------------- INIT --------------------------------//

  printf("Running....\n");
  if (n_models<=0) {
    Serror ("Invalid number of models!\n");
    ABORT_MAIN;
  }
  if ((model = AllocM (PredictorModel*, n_models)) == NULL) {
    Serror ("Could not allocate model array.\n");
    ABORT_MAIN;
  }

  if ((y = AllocM (real, n_models)) == NULL) {
    Serror ("Could not allocate output array!\n");
  }


  for (i=0; i<n_models; i++) {
    y[i] = 0.25f;
    model[i] = NewPredictorModel (jitter*drand48(), variance*drand48(),
                                  bias*(drand48()-0.5));
    if (model[i] == NULL) {
      Serror ("Insufficient memory for all models\n");
      ABORT_MAIN;
    }
  }


  //------------------------------ MAIN LOOP -----------------------------//

  for (t=0; t<input_sequence->length; t++) {
    real sum = 0.0f;
    real mul = 1.0f;
    real imul = 1.0f;
    real tsum = 0.0f;
    real itsum = 0.0f;
    fprintf (output_stream, "%f", input_sequence->x[t]);
    for (i=0; i<n_models; i++) {
      real out =  GetOutput (model[i], input_sequence);
      y[i] = y[i]*0.1f + out*0.9f;
      //      fprintf (output_stream, " %f", out);
    }
    for (i=0; i<n_models; i++) {
      sum += y[i];
      if (y[i]>0.5) {
        tsum = tsum + 1.0f;
        mul = mul*(y[i]+0.01f);
      } else {
        itsum = itsum +1.0f;
        imul = imul*(1.01f-y[i]);
      }
      // moved them up
      //      mul = mul*(0.5f + y[i]);
      //      imul = imul*(1.5f -y[i]);
      // fprintf (output_stream, " %f", y[i]);
    }
    fprintf (output_stream, " %f %f %f %f %f\n",
             sum/((real)n_models), mul, 1.0f - imul,
             tsum/((real)n_models), 1.0f - itsum/((real)n_models));
  }

    
  
  //--------------------------------- EXIT -------------------------------//
 quick_exit:  
  if (exit_condition) {
    fprintf (msgport, "Aborting!!\n\nTrying to clean up the mess...\n");
  } else {
    fprintf (msgport, "Cleaning up\n");
  }

  if (y) {
    FreeM (y);
  }

  if (model) {
    for (i=0; i<n_models; i++) {
      if (model[i]) {
        DeletePredictorModel (model[i]);
      }
    }
    FreeM (model);
  }

  if (input_sequence) {
    FreeSampleBuffer(input_sequence);
    input_sequence = NULL;
  }

  if (output_stream) {
    CloseFile(output_stream);
  }

  if (exit_condition)
    exit (exit_condition);

  fprintf (msgport, "Exiting\n");
  return 0;
} /* main () */
//=========================================================================//
 
//-------------------------------- FUNCTIONS ------------------------------//
// Create a new predictor model
PredictorModel* NewPredictorModel (real jitter, real variance, real bias)
{
  PredictorModel* model = NULL;

  if ((model = AllocM(PredictorModel, 1)) == NULL) {
    Serror ("Could not allocate model\n");
    return NULL;
  }
  assert (variance>=0);
  assert (jitter>=0);

  model->position = 0;
  model->jitter = jitter;
  model->variance = variance;
  model->bias = bias;
  
  model->fpe = (0.5f + bias) * sqrt (variance);
  model->fne = (0.5f - bias) * sqrt (variance);

  model->output = 0.0f;
  printf ("bias: %f    \t variance: %f\n", model->bias, model->variance);
  printf ("FNEG: %f    \t FPOS: %f\n", model->fpe, model->fne);
  if ((model->fpe > 1.0f)||(model->fne>1.0f)) {
    Serror ("Model FPE/FNE over 1.0!\n");
    return NULL;
  }

  return model;
}
// Delete a previously allocated model. It is recommended to use the macro instead.
void FreePredictorModel (PredictorModel* model)
{
  if (!model) {
    Serror ("NULL pointer supplied!\n");
    return;
  }

  FreeM(model);
}

// Get the predictor's output. Cool!
real GetOutput (PredictorModel* model, SAMPLE_BUFFER* buffer)
{
  real y;
  long position;
  
  position = model->position;

  if (drand48() < model->jitter) {
    if (drand48() < 0.5f) {
      position+= (long) (drand48()*10.0f);
    } else {
      position-=(long) (drand48()*10.0f);
    }
  }

  if ((position < 0) || (position >= buffer->length)) {
    return 0.0f;
  }


  model->position = position;
  y = buffer->x[position];

  if (y<(0.5f+model->bias)) {
    if (drand48() < model->variance) {
      y = 1.0f;
    } else {
      y = 0.0f;
    }
  } else {
    if (drand48() < model->variance) {
      y = 0.0f;
    } else {
      y = 1.0f;
    }
  }

  model->output = y;
  
  model->position++;
  return model->output;
}


