/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* $Id: datafile.c,v 1.6 2003/08/27 01:26:37 olethros Exp $ */
/*
  DESCRIPTION
  
  - This simple program is used in conjuction with the DataFile class
  methods, in order to extract symbols from data tables and tag then
  numerically on a per-column basis. Each column is read once and all
  non-numbers are entered into a tag-list. Then conversion occurs.


  SEE ALSO
  - let2num.c
  - datafile.c

*/

#include "datafile.h"


//====================================================================
// ScanDataFile()
//--------------------------------------------------------------------
/* ARGUMENTS

   - FileName: The filename of the file to be scanned

   DESCRIPTION

   - This allocates a DATAFILE structure for file conversion purposes.
   It can handle either Torch format datafiles, or normal matrix ASCII
   datafiles. */

DATAFILE* ScanDataFile(char* FileName) {
  char* StringBuffer = NULL;
  int StringBufferLength = 2048;
  char* err;
  FILE* FileStream;
  DATAFILE* DataFile;

  if (!ReadFile(FileName, FileStream)) {
	Serror ("ERROR opening file\n");
	return NULL;
  }

  if (!(StringBuffer  = AllocM(char, StringBufferLength))) {
	Serror ("Memory allocation failed\n");
	return NULL;
  }
	  
  if (!(DataFile = AllocM(DATAFILE, 1))) {
	Serror("Memory Allocation Failed\n");
	return NULL;
  }
  

  err = fgets(StringBuffer, StringBufferLength, FileStream);
 
  SetDimensionsFromString(DataFile, StringBuffer);
  if (DataFile->rows == 0) {
    do {
      DataFile->rows ++;
      err = fgets(StringBuffer, StringBufferLength, FileStream);
    } while (err);
    rewind(FileStream);
  }

  printf("Data file has %ld columns and %ld rows\n",DataFile->cols, DataFile->rows);

  FreeM(StringBuffer);

  DataFile->InputFile = FileStream;

  return DataFile;

}




//====================================================================
// SetDimensionsFromString
//--------------------------------------------------------------------
/* ARGUMENTS

   - DataFile : The DataFile for which to set the dimensions
   - StringBuffer : This should contain the dimension-describing string

   DESCRIPTION

   - Sets up the file dimensions, given the string

   BUGS AND LIMITATIONS
   
   - This function cannot accurately determine the format at the datafile,
   as it only looks at the provided string.
*/

int  SetDimensionsFromString(DATAFILE* DataFile, char* StringBuffer) {
  int i;
  char* s;

  i = 0;
  s = strtok(StringBuffer, ", ");
  while (s) {
	
	switch(i) {
	case 0:
	  sscanf(s,"%ld", &DataFile->rows);
    break;
	case 1:
	  sscanf(s,"%ld", &DataFile->cols);
    break;
	}
	
	//printf("%d: %s\n", i, s);
	i++;
	s = strtok(NULL, ", ");
  }

  if (i!=2) {
	fprintf (stderr, "First line does not have two columns - Assuming data in tabular format, %d columns\n", i);
	DataFile->cols = i;	
	DataFile->rows = 0;
  } else {
    if ((DataFile->rows>0)&&(DataFile->rows>0)) {
      fprintf(stderr, "First line seems to contain file dimensions - Assuming those work\n");
    } else {
      fprintf(stderr, "First line, though 2-dimensional, contains data - assuming tabular format for file and scanning for number of rows.\n");
      DataFile->cols = i;
      DataFile->rows = 0;
    }
  }
  
  return 0;

}
  


//====================================================================
// TagsToNumbers
//--------------------------------------------------------------------
/* ARGUMENTS

   - DataFile : The DataFile for which to perform the conversion
   - column : The data column to convert

   DESCRIPTION

   - This function can be used to convert a data column from an
   arbitrary format of the type string1,string2,string3... to
   0,1,2. This is useful for programs like Torch, which expect
   class labels to be in the set {0,1,2,..,N-1}.

   BUGS AND LIMITATIONS

   - Text between quotes is not considered a single string when
   performing the conversion.
*/

int TagsToNumbers(DATAFILE* DataFile, long column) {
  LINKEDLIST* list = NULL;
  long i;
  char* StringBuffer = NULL;
  int StringBufferLength = 2048;
  char* s;

  if (!(StringBuffer  = AllocM(char, StringBufferLength))) {
	Serror ("Memory allocation failed\n");
	return -1;
  }


  assert(DataFile);
  assert(DataFile->cols);
  assert(DataFile->rows);
  assert(DataFile->InputFile);
  assert(DataFile->OutputFile);

  printf("Beginning conversion\n");

  fprintf(DataFile->OutputFile, "%ld %ld\n", DataFile->rows, DataFile->cols);
  
  if (!(list = NewList())) {
	Serror ("Error creating list\n");
	return -1;
  }

  for (i=0; i<DataFile->rows; i++) {

	s = fgets(StringBuffer, StringBufferLength, DataFile->InputFile);
	if (s) {
	  char* token;
	  int i = 0;
	  token = strtok(StringBuffer, ", ");

	  while(token) {
		i++;
		if (i==column) {
		  fprintf(DataFile->OutputFile, "%ld ", GetStringMatchingListItem(list, token)->ID);
		} else {
		  fprintf(DataFile->OutputFile, "%s ",token);
		}
	  token = strtok(NULL, ", ");
	  }
	  fprintf(DataFile->OutputFile, "\n");
		
	}else{
	  Serror("s should not be NULL\n");
	  return -1;
	}

  }	

  DeleteList(list);

	return 0;

}
 





///====================================================================
// DisruptSequenceData
//--------------------------------------------------------------------
/* ARGUMENTS

   - DataFile : The input datafile. (The outputfile is left unused, since there are many)
   - prefix : What to prefix the output files with
   - class_column : The column number that contains the class labels
   - seq_column : The column number containing the sequence ID

   DESCRIPTION

   - This function is only used in seq2multi.c - and its output is
   described there in detail.
*/

int DisruptSequenceData(DATAFILE* DataFile, char* prefix, long class_column, long seq_column) {
  LINKEDLIST* list = NULL;
  LINKEDLIST* filelist = NULL;
  LINKEDLIST* complete_list = NULL;
  long i;
  char* StringBuffer = NULL;
  char* BackupBuffer = NULL;
  int StringBufferLength = 2048;
  char* s;
  int MAX_OUT_FILES = 256;
  char t_string[256];
  PACKDATA* pdata = NULL;
  long file_id;
  long classid, seqid;
  long prevseqid=-1;
  long n_input_files = 0;
  long n_target_files = 0;
  PACKDATA** FilePack;
  FILE* cfile;


  /* Prepare the FilePack data */
  if ((FilePack = AllocM(PACKDATA*, MAX_OUT_FILES))==NULL) {
    Serror ("AllocM failed\n");
  }
  for (i=0; i<MAX_OUT_FILES; i++) {
    FilePack[i] = NULL;
  }

  /* Miscellaneous allocations */
  if (!(StringBuffer = AllocM(char, StringBufferLength))) {
    Serror ("Memory allocation failed\n");
    return -1;
  }
  if (!(BackupBuffer = AllocM(char, StringBufferLength))) {
    Serror ("Memory allocation failed!\n");
    return -1;
  }

  assert(DataFile);
  assert(DataFile->cols);
  assert(DataFile->rows);
  assert(DataFile->InputFile);

  printf("Beginning conversion\n");
  if (!(complete_list = NewList())) {
    Serror ("Error creating list\n");
    return -1;
  }
  if (!(filelist = NewList())) {
    Serror ("Error creating list\n");
    return -1;
  }
  if (!(list = NewList())) {
    Serror ("Error creating list\n");
    return -1;
  }


  /* OK - convert files - a bit complicated */
  for (i=0; i<DataFile->rows; i++) {
    s = fgets(StringBuffer, StringBufferLength, DataFile->InputFile);
    strcpy(BackupBuffer, StringBuffer);

    /* While we have something to read */
    if (s) {
      char* token;
      int i = 0;
      token = strtok(StringBuffer, ", ");
      classid = -1;
      seqid = -1;
      /* tokenize the string -unfortunately what we're looking for is
         at the end of the string */
      while(token) {
        i++;
        if (i==class_column) {
          GetStringMatchingListItem(list, token);
          classid = atoi(token);
#if 0
          if (classid<13)
            classid = 0;
          else
            classid = 1;
#endif
        }
        if (i==seq_column) {
          seqid = atoi(token);
        }
        token = strtok(NULL, ", ");
      }

      if ((classid==-1)||(seqid==-1)) {
        Serror ("Unmatched class or sequence id - must be missing columns in the data!\n\nByebye\n");
        return -1;
      }
      
      /* OK, if we move to a new seqid, we can close all our previous
         files and clear the list */
      if (prevseqid==-1)
        prevseqid = seqid;

      if (prevseqid!=seqid) {
        prevseqid=seqid;
        printf("seq : %ld\n",seqid);
        for (i=0; i<MAX_OUT_FILES; i++) {
          if (FilePack[i]) {
            //          printf("Flushing %d to disk\n",i);
            PackDataConvert(FilePack[i]);
            FreePackData(FilePack[i]);
            FilePack[i] = NULL;
          }
        }
        DeleteList(filelist);
        if (!(filelist = NewList())) {
          Serror ("Error creating list\n");
          return -1;
        }
      }

      /* Alright, got the info - find matching output file name - open file if not already open */
      sprintf(t_string, "%sInputs%ld_%ld", prefix, classid, seqid);
      file_id = GetStringMatchingListItem(filelist, t_string)->ID;
      if (file_id>=MAX_OUT_FILES) {
        Serror("Too many files -. Sorry.. :) \n");
        return -1;
      } else {
        pdata = FilePack[file_id];
      }

      if (pdata==NULL) {
        n_input_files++;
        //      printf("New pack data (%ld): %s\n",file_id,t_string);
        pdata = NewPackData(t_string,class_column-1);
        if (pdata) {
          FilePack[file_id] = pdata;
        } else {
          Serror ("Ooops! Could not write file : %s\n",t_string);
          return -1; /* exit uncleanly - bad */
        }
      }

      //    printf("Writing to (%ld)\n",file_id);

      /* We kept a copy of that string, just in case - retokenizing
         until we hit the class column (Assumption: Class column
         contains a string describing the class. No other columns are
         used for that - Presto!)  */
      token = strtok(BackupBuffer, ", ");
      i = 0;
      while(token) {
        i++;
        if (i<class_column) {
          PackDataAddElement(pdata, atof(token));
          token = strtok(NULL, ", ");
        } else {
          //          PackDataRowFinished(pdata);
          sprintf(t_string, "Inputs%ld_%ld Targets%ld_%ld", classid, seqid, classid, seqid);
          GetStringMatchingListItem(complete_list, t_string);
          /* OK - also write the class into the target file */
          n_target_files++;
          sprintf(t_string, "%sTargets%ld_%ld", prefix, classid, seqid);

          if (WriteFile(t_string, cfile)) {
            fprintf(cfile,"1 1\n");
            fprintf(cfile, "%ld\n",classid);
            CloseFile(cfile);
          } else {
            Serror("Ouch! Could not write target description file! How Gross!");
          }
          token = NULL;
        }
      }
      
      
      
    }
  }

  /* All right, split the data into files, but we are not done yet!!
     Make a list of files - useful for brain-dead programs that like
     to have ready-made lists of things. */
  sprintf(t_string,"%sfile.list",prefix);
  /* OK - traverse the list of nodes - should actually put this code
       in LinkedList.c itself but I cannot be bothered right now -
       thank you. */
  if (WriteFile(t_string, cfile)) {
    LISTNODE* curr = complete_list->first; //Yup, ANSI-C allows if in beg. of codeblock
    fprintf(cfile,"%ld 2\n",complete_list->items);
    while (curr) {
      fprintf(cfile, "%s\n", (char*) (curr->data->data));
      curr = curr->next;
    }
  } else {
    Serror ("The file cannot be opened! My KINGDOM for a file-opener!!!\n");
  }

  
  /* Print some info to the user, to make him/her happy */
  printf("Found %ld classes\n",list->items);
  printf("Wrote %ld input files\n",n_input_files);
  printf("%ld total targets\n",n_target_files);
  printf("Wrote 1 file list\n");

	/* Alright, everything is done - Make sure all of our files are
     cleared up */
  for (i=0; i<MAX_OUT_FILES; i++) {
    if (FilePack[i]) {
      //    printf("Flushing %ld to disk\n",i);
      PackDataConvert(FilePack[i]);
      FreePackData(FilePack[i]);
      FilePack[i] = NULL;
      printf("Closing %ld\n",i);
    }
  }
  FreeM(FilePack);

  /* Yeah, clear everything to cover our tracks... */
  DeleteList(complete_list);
  DeleteList(filelist);
  DeleteList(list);

	return 0;

}
 


/* Stupid methods for packed data - I cannot believe that I am writing
   this code! */
void PackDataConvert(PACKDATA* pdata)
{
  LISTNODE* i = NULL;
  FILE* fs = NULL;
  i = pdata->x->first;
  fs  = pdata->FileStream;
  
  if (!fs) {
    Serror("Ooops, no file stream seems to be present here! bad boy!");
    return;
  }

  assert(pdata->rows==pdata->x->items);
  fprintf(fs,"%ld %ld\n",pdata->rows,pdata->cols);
  while(i) {
    float* x = (float *) i->data->data;
    long n;
    for (n=0;n<(pdata->cols-1);n++) {
      fprintf(fs,"%f ",x[n]);
    }
    fprintf(fs,"%f\n",x[pdata->cols-1]);
    i = i->next;
  }
   
}
PACKDATA* NewPackData(char* filename, long n_columns)
{
  PACKDATA* pdata = NULL;
  pdata = AllocM(PACKDATA, 1);
  if (pdata==NULL) {
    Serror("Cannot allocate new pack data\n");
    return NULL;
  }
  
  pdata->x = NewList();
  if (!pdata->x) {
    Serror("Cannot allocate list structure for the packed data\n");
  }

  assert(filename);
  if (!(WriteFile(filename, pdata->FileStream))) {
    Serror("Could not write to file %s\n",filename);
  }
  pdata->curr_column = 0;
  pdata->rows = 0;
  pdata->cols = n_columns;
  pdata->temp = NULL;
  return pdata;
  
}
void FreePackData(PACKDATA* pdata)
{
  DeleteList(pdata->x);
  CloseFile(pdata->FileStream);
  FreeM(pdata);
}

void PackDataAddElement(PACKDATA* pdata, float token)
{
  float* x;
  /* OK, this is a cool trick - if the pointer to the current buffer
     is NULL then it means that either the previous one has been
     already filled, or that we are just starting out. In any case, we
     need to create a new buffer to hold the data and put it in the
     list of rows. Note we could also just look at the list's current
     item, but this is faster - besides, I don't want to trust the
     list, cuz something else might play around with the current
     item.*/

  if (pdata->temp==NULL) {
    pdata->temp = NewBuffer(sizeof(float)*pdata->cols);
    if (!pdata->temp) {
      Serror("Cannot store any more data!\n");
      return;
    }
    if (AddItem(pdata->x, pdata->temp)==NULL) {
      Serror("Cannot append to the list of data! Tough..\n");
      return;
    }
    pdata->rows++;
  }


  /* Now, the next part of this exercise is very simple. Yeah, I know
     I should've provided a method to get the data from it, but who gives?
     Anyway, - Just store the data - That's all.

     Oh, and check when we're off the boundary to set up the reset conditions
  */
  x = (float*) pdata->temp->data;
  x[pdata->curr_column] = token;
  pdata->curr_column++;
  if (pdata->curr_column==pdata->cols) {
    pdata->curr_column = 0;
    pdata->temp = NULL;
  }
  
  /* Now, that's all - redundant return follows */

  return;
}  





//====================================================================
// readFileList() - reads a list of file pairs
//--------------------------------------------------------------------
int readFileList(char* file, FilePair* pairs) {
  FILE* fs;
  char** input_file;
  char** target_file;
  long n_files = 0;
  long n_dim = 0;

  pairs->input_files = NULL;
  pairs->target_files = NULL;
  pairs->n_files = 0;
  pairs->n_dim = 0;

  //  fprintf(stderr,"Reading paired list: %s\n",file);

  if ((fs=fopen(file,"r"))) {
    char tb[256];
    char* s;
    int j,i;

    s = fgets(tb, 256, fs);
    s = strtok(tb, ", ");
    i = 0;
    while (s) {

      switch(i) {
      case 0: sscanf(s, "%ld", &n_files);break;
      case 1: sscanf(s, "%ld", &n_dim);break;
      }
      i++;
      s = strtok(NULL, ", ");
    }
    //    fprintf(stderr,"%ld %ld-tuples of files\n",n_files,n_dim);
    input_file =  AllocM(char*, n_files);
    target_file = AllocM(char*, n_files);
    assert(input_file);
    assert(target_file);

    //    fprintf(stderr,"Scanning..\n");
    j = 0;
    s = fgets(tb, 256, fs);
    while (s) {
      char* token;
      int i = 0;
      int len;
      token = strtok(tb, ", \n");
      while (token) {
        len = strlen(token) + 1;
        switch (i) {
        case 0:
          input_file[j] = AllocM(char, len);
          strcpy(input_file[j],token);
          break;
        case 1:
          target_file[j] = AllocM(char, len);
          strcpy(target_file[j],token);
          break;
        default:
          Serror ("reached column %d, containing [%s] but cannot handle more than 2-tuples\n",i,token);
          return -1;
        } /* switch (i) */
        i++;
        token = strtok(NULL, ", \n");
      } /* while (token) */
      j++;
      s = fgets(tb, 256, fs);
    } /* while (s) */
    fclose(fs);
  } else { /* if (fopen()) */        
    Serror ("Could not open file\n");
    return -1;
  }
  
  pairs->input_files = input_file;
  pairs->target_files = target_file;
  pairs->n_files = n_files;
  pairs->n_dim = n_dim;
  
  return 0;
} 
 
//====================================================================
// readFileList() - reads a list of files
//--------------------------------------------------------------------
char** readFileList(char* file) {
   FILE* fs;
   char** input_file;
   long n_files = 0;
   long n_dim = 0;
   
   printf("Reading single list: %s\n",file);
   
   if ((fs=fopen(file,"r"))) {
     char tb[256];
     char* s;
    int j,i;

    s = fgets(tb, 256, fs);
    s = strtok(tb, ", ");
    i = 0;
    while (s) {
      switch(i) {
      case 0: sscanf(s, "%ld", &n_files);break;
      case 1: sscanf(s, "%ld", &n_dim);break;
      }
      i++;
      s = strtok(NULL, ", ");
    }
    input_file =  AllocM(char*, n_files);
    assert(input_file);

    j = 0;
    s = fgets(tb, 256, fs);
    while (s) {
      char* token;
      int i = 0;
      int len;
      token = strtok(tb, ", \n");
      while (token) {
        len = strlen(token) + 1;
        switch (i) {
        case 0:
          input_file[j] = AllocM(char, len);
          strcpy(input_file[j],token);
          break;
        default:
          fprintf (stderr, "reached column %d, containing [%s] but cannot handle more than 1-tuples\n",i,token);
          return NULL;
        } /* switch (i) */
        i++;
        token = strtok(NULL, ", \n");
      } /* while (token) */
      j++;
      s = fgets(tb, 256, fs);
    } /* while (s) */
    fclose(fs);
   } else {
     Serror ("Could not open file");
     return NULL;
   }
   return input_file;
} 

