/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
const char *PSR_VERSION =  "VER: $Id: psr.c,v 1.7 2003/09/03 11:36:50 olethros Exp $";


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "debug.h"
#include "psr.h"
#include "modulator.h"
#include <assert.h>

//==========================================================
// CreateSine(long,....) 
//----------------------------------------------------------
/* 
   Creates a sine wave.

   ARGUMENTS
   - length : the length of the sinewave (in samples)
   - r : the sampling rate
   - f : the frequency of the sinewave
   - a : the amplitude of the sinewave

   DESCRIPTION
   - Allocates memory for and Creates a sinewave of the specified length, amplitude and frequency.
   
   RETURNS

   A pointer to the SAMPLE_BUFFER containing the new sample

   SEE ALSO

   CreateSine(float,...)
*/
   
SAMPLE_BUFFER* CreateSine(long length, float r, float f, float a) {
  register long i;
  register float T;
  register float* x;
  register float u = 0.0f;
  register float pi2;

  assert(r);
  assert(f);
  assert(a);
  assert(length>=0);

  pi2 = 2.0f*M_PI; // <<--- M_PI Should be defined

  T = pi2*f/r;
  T = fabs(T); // <<--- accept negative values, but ignore them :)

  SAMPLE_BUFFER* buf = NewSampleBuffer(length, 0); // <<--- no need to clear space since we are writing over it
  assert(buf->length==length);
  x = buf->x;
  for(i=0; i<length; i++) {
    *x++ = a*sin(u);
    u += T;
    /* We need to do a wraparound because of accuracy loss
       (especially for long samples) */
    if (u>pi2) u -= pi2; // <<--- Just one wraparound check needed (T>0)
  }
  buf->rate = (long) r;
  return buf;
}

//==========================================================
// CreateSine(float,....) 
//----------------------------------------------------------
/* 
   Creates a sinewave of length specified in seconds.

   ARGUMENTS
   - length : the length of the sinewave (in seconds)

   SEE ALSO
   - Look at the description of CreateSine(long,...) for details

*/
SAMPLE_BUFFER* CreateSine(float length, float r, float f, float a) {
  long l;
  if (length<=0) {
    Serror("Negative length???\n");
    return NULL;
  }

  /* calculate length in samples */
  l = (long)(r*length); // <<--- easy, huh? :)
  
  return (CreateSine(l, r, f, a));
}





//==========================================================
// CreatePulse(long...)
//----------------------------------------------------------
/* 
   Creates a pulse wave.

   ARGUMENTS
   - length : the length of the pulsewave (in samples)
   - r : the sampling rate
   - f : the frequency of the pulsewave
   - a : the amplitude of the pulsewave

   DESCRIPTION
   - Allocates memory for and Creates a pulsewave of the specified length, amplitude and frequency.
   
   RETURNS

   A pointer to the SAMPLE_BUFFER containing the new sample

   SEE ALSO

   CreatePulse(float,...)
*/
   
SAMPLE_BUFFER* CreatePulse(long length, float r, float f, float a) {
  register long i;
  register float T;
  register float* x;
  register float u = 0.0f;
  register float pi2;

  assert(r);
  assert(f);
  assert(a);
  assert(length>=0);

  pi2 = 2.0f*M_PI; // <<--- M_PI Should be defined

  T = pi2*f/r;
  T = fabs(T); // <<--- accept negative values, but ignore them :)

  SAMPLE_BUFFER* buf = NewSampleBuffer(length, 0); // <<--- no need to clear space since we are writing over it
  assert(buf->length==length);
  x = buf->x;
  for(i=0; i<length; i++) {
    if (sin(u)>=0.0f) {
      *x++ = a;
    } else {
      *x++ = -a;
    }

    u += T;
    /* We need to do a wraparound because of accuracy loss
       (especially for long samples) */
    if (u>pi2) u -= pi2; // <<--- Just one wraparound check needed (T>0)
  }
  buf->rate = (long) r;
  return buf;
}




  
//==========================================================
// CreatePulse(float,....) 
//----------------------------------------------------------
/* 
   Creates a pulse of length specified in seconds.
*/
SAMPLE_BUFFER* CreatePulse(float length, float r, float f, float a) {
  long l;
  if (length<=0) {
    Serror("Negative length???\n");
    return NULL;
  }

  /* calculate length in samples */
  l = (long)(r*length); // <<--- easy, huh? :)
  
  return (CreatePulse(l, r, f, a));
}

  
//==========================================================
// CreateSaw(long...)
//----------------------------------------------------------
/* 
   Creates a saw wave.

   ARGUMENTS
   - length : the length of the sawwave (in samples)
   - r : the sampling rate
   - f : the frequency of the sawwave
   - a : the amplitude of the sawwave

   DESCRIPTION
   - Allocates memory for and Creates a sawwave of the specified length, amplitude and frequency.
   
   RETURNS

   A pointer to the SAMPLE_BUFFER containing the new sample

   SEE ALSO

   CreateSaw(float,...)
*/
   
SAMPLE_BUFFER* CreateSaw(long length, float r, float f, float a) {
  register long i;
  register float T;
  register float* x;
  register float u = 0.0f;
  register float pi2;

  assert(r);
  assert(f);
  assert(a);
  assert(length>=0);

  pi2 = 2.0f*M_PI; // <<--- M_PI Should be defined

  T = pi2*f/r;
  T = fabs(T); // <<--- accept negative values, but ignore them :)

  SAMPLE_BUFFER* buf = NewSampleBuffer(length, 0); // <<--- no need to clear space since we are writing over it
  assert(buf->length==length);
  x = buf->x;
  for(i=0; i<length; i++) {
    *x++ = a*(2*u-pi2)/pi2;
    u += T;
    /* We need to do a wraparound because of accuracy loss
       (especially for long samples) */
    if (u>pi2) u -= pi2; // <<--- Just one wraparound check needed (T>0)
  }

  buf->rate = (long) r;
  return buf;
}




  
//==========================================================
// CreateSaw(float,....) 
//----------------------------------------------------------
/* 
   Creates a saw of length specified in seconds.
*/
SAMPLE_BUFFER* CreateSaw(float length, float r, float f, float a) {
  long l;
  if (length<=0) {
    Serror("Negative length???\n");
    return NULL;
  }

  /* calculate length in samples */
  l = (long)(r*length); // <<--- easy, huh? :)
  
  return (CreateSaw(l, r, f, a));
}

  
//==========================================================
// AddSine()
//----------------------------------------------------------
SAMPLE_BUFFER* AddSine(SAMPLE_BUFFER* buf, float r, Modulator* f, Modulator* a) {
  register long i;
  register float T;
  register float* x;
  register float u = 0.0f;
  register float pi2;
  long length = buf->length;

  assert(r);
  assert(f);
  assert(a);
  assert(length>=0);

  pi2 = 2.0f*M_PI; // <<--- M_PI Should be defined


  x = buf->x;
  for(i=0; i<length; i++) {
    *x++ = *x + sin(u) * MODFN (a, i);
    T = MODFN(f,i) * pi2/r;
    T = fabs(T);
    u += T;
    /* We need to do a wraparound because of accuracy loss
       (especially for long samples) */
    if (u>pi2) u -= pi2; // <<--- Just one wraparound check needed (T>0)
  }

  return buf;
}


			    
//==========================================================
// AddUniformNoise()
//----------------------------------------------------------
int AddUniformNoise(SAMPLE_BUFFER* sample, float amplitude) {
  long l;
  float* x;
  if (CheckBuffer(sample)) return -1;
  
  x = sample->x;
  for (l=0;l<sample->length;l++) {
    *x += (drand48()-0.5f)*amplitude;
    x++;
  }
  
  
  return 0;
}


			    
//==========================================================
// AddNoise()
//----------------------------------------------------------
int AddUniformNoise(SAMPLE_BUFFER* sample, Modulator* a) {
  long l;
  float* x;
  if (CheckBuffer(sample)) return -1;
  
  x = sample->x;
  for (l=0;l<sample->length;l++) {
    float xx = (float) l;
    *x += (drand48() - 0.5f) * (MODFN (a, xx));
    x++;
  }
  
  
  return 0;
}


//==========================================================
// CreatePseudoRandomSequence()
//----------------------------------------------------------
/* 
   Creates a pseudorandom sequence

   ARGUMENTS
   - length : the length of the sequence (in samples)
   - p : the probability of a '1' in the sequence


   DESCRIPTION
   - Allocates memory for and Creates a pseudorandom of the specified length and with the specified probabilities.
   
   RETURNS

   - A pointer to the SAMPLE_BUFFER containing the new sample

   SEE ALSO

   - CreateSine(float,...)
*/
   
SAMPLE_BUFFER* CreatePseudoRandomSequence(long length, float p) {
  register long i;
  register float* x;

  assert(length>=0);

  if ((p<0.0f)||(p>1.0f)) {
    Swarning("The probability of a '1' in the sequence should be a number between 0 and 1. Continuing anyway");
  }

  SAMPLE_BUFFER* buf = NewSampleBuffer(length, 0); // <<--- no need to clear space since we are writing over it
  assert(buf->length==length);
  x = buf->x;

  for(i=0; i<length; i++) {
    if (drand48()<p) {
      *x++ = 1.0f;
    }else{
      *x++ = 0.0f;
    }
  }

  return buf;
}

//==========================================================
// KernelFilter()
//----------------------------------------------------------
/* 
   ARGUMENTS
   - sample : The sample to be filtered
   - beta : Filter parameter
   - gamma : Filter parameter

   DESCRIPTION

   - Filters a sample buffer with an online approximation of a
     gaussian kernel.
   
   RETURNS

   - A pointer to the SAMPLE_BUFFER containing the sample, or NULL if
     the operation failed.

   SEE ALSO

   - LowPassFilter()
*/
SAMPLE_BUFFER* KernelFilter (SAMPLE_BUFFER* sample, Modulator* beta, Modulator* gamma)
{
  long l;
  float* x_in;
  float x = 0;
  float ix = 0;
  float dx = 0;
  float y = 0;

  if (CheckBuffer(sample))
    return NULL;

  x_in = sample->x;

  for (l=0;l<sample->length;l++) {
    y = *x_in;
    dx = (1.0 - MODFN (beta, l))*(ix-x);
    ix = ix * MODFN (beta, l) + y;
    x = x * MODFN (gamma, l) + (1.0 - MODFN (gamma, l)) * ix;
    *x_in += x;
    x_in++;
  }
  
  return sample;
}


//==========================================================
// Oscillator()
//----------------------------------------------------------
/* 
   Oscillates with the input
   
   ARGUMENTS
   sample - A pointer to the SAMPLE_BUFFER containing the new sample,
   if a new sample is desired.
   r - The sampling rate
   f - The oscillator's natural frequency
   d - The damping value for the oscillator
*/
SAMPLE_BUFFER* Oscillator (SAMPLE_BUFFER* sample, float r, float f, float d)
{
  long l;
  float T;
  float y;
  float dy;
  float* x;
  float pi2;
  assert(r);
  assert(f);

  printf ("Oscillator...\n");
  pi2 = 2.0f*M_PI; // <<--- M_PI Should be defined

  T = f/r;
  T = fabs(T); // <<--- accept negative values, but ignore them :)

  /* Now, create a simple filter */
  assert (sample);
  assert (sample->x);
  x = sample->x;
  y = 0.0f;
  dy = 0.0f;
  for (l=0; l < sample->length; l++) {
    //for (l=0; l < 100; l++) {
    dy +=  - T*y - 0.1*d*dy;
    y += 0.01*(*x) + 0.1*dy;
    *x++ = y;
    //    printf ("%f\n",y);
  }
  return sample;
}

//==========================================================
// VolumeRamp
//----------------------------------------------------------
/* 
   ARGUMENTS
   - sample : The sample to be filtered
   - r : sample rate
   - start : starting time, in seconds.
   - end : ending time, in seconds
   - boost_a : boosting at start of operation
   - boosg_b : boosting at end of operation

   DESCRIPTION

   - amplifies a sample's amplitude within the time range [start, end]
   by the amount X, where X is interpolated from linearly from boost_a
   to boost_b.
   
   RETURNS

   - A pointer to the SAMPLE_BUFFER containing the sample, or NULL if
     the operation failed.

   SEE ALSO

   - Amplify()
*/
SAMPLE_BUFFER* VolumeRamp (SAMPLE_BUFFER* sample, TimeStamp start, TimeStamp end, float boost_a, float boost_b)
{
  float r = (float) sample->rate;
  float s = ((float) start) / r;
  float e = ((float) end) / r;

  return VolumeRamp (sample, r, s, e, boost_a, boost_b);

}
SAMPLE_BUFFER* VolumeRamp (SAMPLE_BUFFER* sample, float r, float start, float end, float boost_a, float boost_b)
{
  long x_start;
  long x_end;
  long i;
  float l;
  float* x;
  float n;
  float dl;
  float fs = r*start;
  float fe = r*end;
  x_start = (long) fs;
  x_end = (long) fe;
  
  if ((sample==NULL)||(sample->x == NULL)) {
    return NULL;
  }


  if (x_start<0) {
    Serror ("Starting at negative time");
    return NULL;
  }
  if (x_start>sample->length) {
    Serror ("Starting at super time");
    return NULL;
  }
  
  x = &sample->x[x_start];
  if (x_end>sample->length) {
    Swarning ("Truncating volume ramp");
    x_end = sample->length;
  };

  l = (float) (x_end - x_start);

  n = 0.0f;
  dl = 1.0/(l);

  for (i=x_start; i<x_end; i++, x++, n+=dl) {
    *x = *x * ((n)*boost_b + (1.0f-n)*boost_a);
  }
  return sample;
}


//==========================================================
// LowPassFilter()
//----------------------------------------------------------
/* 
   ARGUMENTS

   - sample : The sample to be filtered

   - damp : The dampening filter parameter - the closer it is to 1.0
     the more high frequencies are cutoff.

   DESCRIPTION

   - A simple, small, fast, IIR filter for low pass filtering.
   
   RETURNS

   - A pointer to the SAMPLE_BUFFER containing the sample, or NULL if
     the operation failed.

   SEE ALSO

   - LowPassFilter()
*/
SAMPLE_BUFFER* LowPassFilter (SAMPLE_BUFFER* sample, float damp)
{
  long l;
  float* x;
  float y;
  if (CheckBuffer(sample))
    return NULL;

  x = sample->x;

  if ((damp < 0.0f) || (damp > 1.0f)) {
    Swarning ("float damp should lie in [0,1]");
  }
  y = 0.0f;
  for (l=0; l<sample->length; l++) {
    y = (damp) * (y) + (1.0f - damp) * (*x);
    *x++ = y;
  }
  return sample;
}


SAMPLE_BUFFER* IIRFilter (SAMPLE_BUFFER* sample, Filter* parameters)
{
  long l;
  int An;
  int Bn;
  float* x = NULL; /* Store input and */
  float* y = NULL; /* output history */
  float* a;
  float* b;
  int cnt_a;
  int cnt_b;
  if (CheckBuffer(sample))
    return NULL;

  assert (parameters);

  An = parameters->a->n;
  Bn = parameters->b->n;
  a = parameters->a->x;
  b = parameters->b->x;

  assert (An);
  assert (Bn);
  assert (a);
  assert (b);

  y = AllocM (float, Bn);
  if (y==NULL) {
    Serror ("Allocation failed\n");
    return NULL;
  }
  for (l=0; l<Bn; l++) {
    y[l] = 0.0f;
  }
  
  cnt_a = 0;
  cnt_b = 0;

  for (l=0; l<sample->length; l++) {
    int i;
    int j = cnt_b;
    float u = 0.0f;
    x = &sample->x[l];
    for (i=(An-1); i>=0; i--) {
      if ((l-i)>=0) {
        u += x[i]*a[i];
      }
    }
    for (i=(Bn-1); i>=0; i--) {
      u += y[j]*a[i];
      j++; if (j==Bn) j=0;
    }
    cnt_b ++;
    if (cnt_b==Bn) {
      cnt_b=0;
    }
    y[cnt_b] = u;
  }
  return sample;
}


